"""
This is the BAGUETTE framework. BAGUETTE stands for Behavioral Analysis Graph Using Execution Traces Towards Explanability.
BAGUETTE is a heterogeneous graph data structure used to represent the behavior of malware samples.
"""


"""
This module defines the Vertices and Edges that are related to registry activity.
"""

from ...logger import logger
from typing import Any, Optional, Union
from ..colors import Color
from ..graph import Arrow, Edge, UniqueVertex, Vertex
from .execution import Call, CallHandler, Process
from .network import Host
from .filesystem import File, Directory
from ..utils import chrono
from ..build import BuildingPhase

__all__ = ["Key", "KeyEntry", "HasEntry", "HasSubKey"]





logger.debug("Loading registry library.")

class Key(UniqueVertex):

    """
    A registry key vertex. Represents a key without its associated value(s).
    """

    __slots__ = {
        "__name" : "The local name of the registry key int the registry tree."
    }

    __pickle_slots__ = {
        "name"
    }

    def __init__(self, *, c: Color | None = None, parent: Optional["Vertex"] = None) -> None:
        super().__init__(c=c, parent=parent)
        self.__name : str = ""

    @property
    def name(self) -> str:
        """
        The (local) name of the registry key.
        """
        return self.__name
    
    @name.setter
    def name(self, value : str):
        if not isinstance(value, str):
            raise TypeError("Expected str, got " + repr(type(value).__name__))
        if "\\" in value:
            raise ValueError("'\\' not permitted in a registry key name")
        if len(value) > 255:
            raise ValueError("Key name is too long to be allowed by registry")
        self.__name = value.lower().title().replace("\x00", "\uFFFD")
        if not self.printable:
            from ..config import ColorMap
            from ...logger import logger
            logger.info("Got an unprintable registry key.")
            self.color = ColorMap.get_config().registry.UnprintableKey
    
    @property
    def path(self) -> str:
        """
        The global key name (path from root key and all sub-keys to reach this key included).
        """
        if not self.parent_key:
            return self.name
        return self.parent_key.path + "\\" + self.name

    @property
    def printable(self) -> bool:
        """
        True if the (relative) key name is printable.
        """
        return self.name.isprintable() and "\uFFFD" not in self.name

    @property
    def parent_key(self) -> Union["Key", None]:
        """
        The parent key of this registry key if any (None for root keys).
        """
        for e in self.edges:
            if isinstance(e, HasSubKey) and e.destination is self and isinstance(e.source, Key):
                return e.source
    
    @property
    def entries(self) -> list["KeyEntry"]:
        """
        All the known entries of this key.
        """
        return [e.destination for e in self.edges if isinstance(e, HasEntry)]
    
    @property
    def label(self) -> str:
        """
        The label of this node.
        """
        return "Key " + self.name





class HasSubKey(Arrow):

    """
    This kind of arrow indicates that a key has a sub-key.
    """

    label : str = ""

    source : Host | Key
    destination : Key





class KeyEntry(UniqueVertex):

    """
    A registry key entry vertex. This represents a value (with name, type and actual value) that a registry key has.
    """

    __slots__ = {
        "__name" : "The name of the registry key entry",
        "__type" : "The type of value for this entry",
        "__value" : "The actual value stored in this entry"
    }

    __pickle_slots__ = {
        "name",
        "reg_type",
        "value"
    }

    def __init__(self, *, c: Color | None = None, parent: Optional["Vertex"] = None) -> None:
        super().__init__(c=c, parent=parent)
        self.__name : str = ""
        self.__type : str = ""
        self.__value = None
        self.reg_type = "REG_NONE"

    @property
    def name(self) -> str:
        """
        The name of this registry key entry.
        """
        return self.__name
    
    @name.setter
    def name(self, value : str):
        if not isinstance(value, str):
            raise TypeError("Expected str, got " + repr(type(value).__name__))
        if len(value) > 16383:
            raise ValueError("Registry key value name is too long.")
        self.__name = value
    
    @property
    def reg_type(self) -> str:
        """
        The type of data stored in the registry key entry.
        """
        return self.__type

    @reg_type.setter
    def reg_type(self, value : str):
        if not isinstance(value, str):
            raise TypeError("Expected str, got " + repr(type(value).__name__))
        if value not in {"REG_BINARY", "REG_DWORD", "REG_DWORD_LITTLE_ENDIAN", "REG_DWORD_BIG_ENDIAN", "REG_EXPAND_SZ", "REG_LINK", "REG_MULTI_SZ", "REG_NONE", "REG_QWORD", "REG_QWORD_LITTLE_ENDIAN", "REG_SZ"}:
            raise ValueError("Not a valid type fr registry key : " + str(value))
        self.__type = value
        if self.__type == "REG_NONE":
            self.__value = None
    
    def __setstate__(self, state: dict[str, Any]):
        state = {"reg_type" : state["reg_type"]} | state        # Attribute "reg_type" should be loaded before attribute "value"
        return super().__setstate__(state)
            
    @property
    def value(self) -> bytes | int | str | list[str] | None:
        match self.__type:
            case "REG_BINARY":
                return self.__value
            case "REG_QWORD_LITTLE_ENDIAN" | "REG_DWORD_BIG_ENDIAN" | "REG_DWORD" | "REG_DWORD_LITTLE_ENDIAN" | "REG_QWORD":
                return self.__value
            case "REG_LINK":
                return self.__value
            case "REG_NONE":
                return None
            case "REG_EXPAND_SZ":
                return self.__value
            case "REG_SZ":
                return self.__value
            case "REG_MULTI_SZ":
                return self.__value
            case _:
                raise RuntimeError("Got a KeyEntry with an unknown type : '{}'".format(self.__type))

    @value.setter
    def value(self, val : bytes | int | str | list[str] | None):
        match self.__type:
            case "REG_BINARY":
                if not isinstance(val, bytes):
                    raise TypeError("Expected bytes for key entry of type '{}'".format(self.__type))
            case "REG_QWORD_LITTLE_ENDIAN" | "REG_DWORD_BIG_ENDIAN" | "REG_DWORD" | "REG_DWORD_LITTLE_ENDIAN" | "REG_QWORD":
                if not isinstance(val, int):
                    raise TypeError("Expected int for key entry of type '{}'".format(self.__type))
            case "REG_LINK":
                if not isinstance(val, str):
                    raise TypeError("Expected str for key entry of type '{}'".format(self.__type))
            case "REG_NONE":
                if val != None:
                    raise TypeError("Expected None for key entry of type '{}'".format(self.__type))
            case "REG_EXPAND_SZ":
                if not isinstance(val, str):
                    raise TypeError("Expected str for key entry of type '{}'".format(self.__type))
            case "REG_SZ":
                if not isinstance(val, str):
                    raise TypeError("Expected str for key entry of type '{}'".format(self.__type))
            case "REG_MULTI_SZ":
                if not isinstance(val, list):
                    raise TypeError("Expected list for key entry of type '{}'".format(self.__type))
                for vi in val:
                    if not isinstance(vi, str):
                        raise TypeError("Expected list of str, got an element of type " + repr(type(vi).__name__))
            case _:
                raise RuntimeError("Got a KeyEntry with an unknown type : '{}'".format(self.__type))
        self.__value = val
    
    def process_value(self, val : str | int):
        match self.__type:
            case "REG_LINK" | "REG_EXPAND_SZ" | "REG_SZ":
                if not isinstance(val, str):
                    raise TypeError("Expected str, got " + repr(type(val).__name__))
                self.__value = val
            case "REG_BINARY":
                if not isinstance(val, str):
                    raise TypeError("Expected str, got " + repr(type(val).__name__))
                self.__value = val.encode()
            case "REG_MULTI_SZ":
                if not isinstance(val, str):
                    raise TypeError("Expected str, got " + repr(type(val).__name__))
                if val.endswith("\x00"):
                    v = val[:-1]
                else:
                    v = val
                self.__value = v.split("\x00")
            case "REG_QWORD_LITTLE_ENDIAN" | "REG_DWORD_BIG_ENDIAN" | "REG_DWORD" | "REG_DWORD_LITTLE_ENDIAN" | "REG_QWORD":
                if not isinstance(val, int):
                    raise TypeError("Expected int, got " + repr(type(val).__name__))
                self.__value = val
            case "REG_NONE":
                if not isinstance(val, str):
                    raise TypeError("Expected str, got " + repr(type(val).__name__))
                if val:
                    raise ValueError("Expected empty string, got " + str(val))
                self.__value = None
            case _:
                raise RuntimeError("Got a KeyEntry with an unknown type : '{}'".format(self.__type))
                    
    @property
    def key(self) -> Key:
        """
        The registry Key that this entry is part of.
        """
        for e in self.edges:
            if isinstance(e, HasEntry):
                return e.source
        raise RuntimeError("Found a KeyEntry that has not been affected to a Key")





class HasEntry(Edge):

    """
    This kind of edge indicates that a registry key has one registry entry.
    """

    label : str = ""

    source : Key
    destination : KeyEntry





class Handle(UniqueVertex):

    """
    A handle vertex. Represents a registry key handle, used when a program opens a registry key.
    """

    def __init__(self, *, c: Color | None = None, parent: Optional[Vertex] = None) -> None:
        super().__init__(c=c, parent=parent)
    
    @property
    def key(self) -> Key:
        """
        Returns the key node that this handle is working on.
        """
        for e in self.edges:
            if isinstance(e, UsesKey):
                return e.destination
        raise RuntimeError("Key handle with no attached key")
            




class UsesKey(Edge):

    """
    This kind of edge indicates that a handle uses a registry key.
    """

    label : str = ""

    source : Handle
    destination : Key





class HasHandle(Edge):

    """
    This kind of edge indicates that a process owns a registry key handle.
    """

    label : str = ""

    source : Process
    destination : Handle





class Discovered(Arrow):

    """
    This kind of arrow indicates that a key handle discovered a sub-key throught enumeration.
    """

    source : Handle
    destination : Key





class QueriesEntry(Arrow):

    """
    This kind of arrow indicates that a key entry was queried by a key handle. 
    """

    source : KeyEntry
    destination : Handle





class SetsEntry(Arrow):

    """
    This kind of arrow indicates that a key handle set the value of key entry.
    """

    source : Handle
    destination : KeyEntry





class DeletesEntry(Edge):

    """
    This kind of edge indicates that a key handle deleted a key entry.
    """

    source : Handle
    destination : KeyEntry





class ChangesTowards(Arrow):

    """
    This kind of arrow indicates that a key entry's content changed to another.
    """

    label : str = ""

    source : KeyEntry
    destination : KeyEntry





class ReferencesFileSystem(Edge):
    
    """
    This kind of edge indicates that a key's entry references a file or a folder.
    """

    source : KeyEntry
    destination : File | Directory





_existing_keys : dict[str, Key] = {}
_inverted_handles : dict[str, Handle] = {}
_active_handles : dict[Handle, str] = {}
_existing_entries : dict[Key, dict[str, KeyEntry]] = {}
_deleted_entries : dict[Key, dict[str, KeyEntry]] = {}

@chrono
def create_key_tree(key : str) -> Key:
    """
    Creates all the missing keys in leading to the final key and returns the leaf Key node.
    """
    from ..graph import find_or_create
    from .network import Host

    k = None
    last = find_or_create(Host, domain = "host")[0]
    for name in key.replace("\x00", "\uFFFD").split("\\"):
        name = name.lower().title()
        k = None
        for e in last.edges:
            if isinstance(e, HasSubKey) and e.source is last and e.destination.name == name:
                k = e.destination
                break
        if not k:
            k = Key()
            k.name = name
            HasSubKey(last, k)
            _existing_keys[k.path.lower()] = k
        last = k
    
    if not k:
        raise RuntimeError("Trying to create an unnamed registry key!")

    return k

@chrono
def split_key_and_entry_names(path : str) -> tuple[str, str]:       # Might evolve later
    """
    Splits the given key entry path into the path of the last subkey and the name of the entry.
    """
    i = None
    for j, c in enumerate(reversed(path)):
        if c == "\\":
            i = -j - 1
            break
    if i == None:
        raise ValueError("Expected at least one '\\', got " + repr(path))
    return path[:i], path[i + 1:]

@chrono
def find_last_key_entry(key : Key, name : str) -> KeyEntry | None:
    """
    Given a registry key and an entry name, returns the last entry known with this name.
    """
    if key in _existing_entries and name in _existing_entries[key]:
        return _existing_entries[key][name]

    



@chrono
def integrate_key_opening(c : Call):
    """
    Creates a new Key node if necessary when a key is opened.
    """
    from ...logger import logger
    if c.status == 1:
        logger.debug("Opening registry key.")
        if c.arguments.regkey.lower() in _existing_keys:
            k = _existing_keys[c.arguments.regkey.lower()]
        else:
            k = create_key_tree(c.arguments.regkey)
        h = Handle()
        UsesKey(h, k)
        HasHandle(c.thread.process, h)
        if c.arguments.key_handle.lower() in _inverted_handles:
            logger.warning("Opening an already existing registry key handle.")
            return
        _inverted_handles[c.arguments.key_handle.lower()] = h
        _active_handles[h] = c.arguments.key_handle.lower()
        
@chrono
def integrate_key_enumeration(c : Call):
    """
    Finds the enumerated Key and sub-Key. Creates the sub-Key if it does not exist. Links the sub-Key to the Handle.
    """
    from ...logger import logger
    if c.status == 1:
        logger.debug("Enumerating from registry key.")
        k : Key
        sk : Key
        if c.arguments.key_handle.lower() not in _inverted_handles:
            logger.warning("Trying to enumerate a registry key with no known handle.")
            return
        h = _inverted_handles[c.arguments.key_handle.lower()]
        k = h.key
        if c.name != "NtEnumerateKey":
            sk = create_key_tree(k.path + "\\" + c.arguments.key_name)
        else:
            sk = create_key_tree(k.path + "\\" + c.arguments.buffer.encode("utf-8").decode("utf-16", errors="replace"))     # Cuckoo tried its best...which is quite bad!
        Discovered(h, sk)

@chrono
def integrate_key_closing(c : Call):
    """
    Closes a Handle associated with a Key.
    """
    from ...logger import logger
    if c.status == 1:
        logger.debug("Closing key handle.")
        if c.arguments.key_handle.lower() not in _inverted_handles:
            logger.warning("Trying to close unseen key handle.")
            return
        h = _inverted_handles.pop(c.arguments.key_handle.lower())
        _active_handles.pop(h)

@chrono 
def integrate_key_deleting(c : Call):
    """
    Marks a registry Key as deleted.
    """
    from ...logger import logger
    from ..config import ColorMap
    if c.status == 1:
        logger.debug("Deleting registry key.")
        if c.arguments.key_handle.lower() not in _inverted_handles:
            logger.warning("Trying to delete key from unseen handle.")
            return
        k = create_key_tree(c.arguments.regkey)
        k.color = ColorMap.get_config().Deleted
        _existing_keys.pop(k.path.lower())
        if k not in _deleted_entries:
            _deleted_entries[k] = {}
        if k in _existing_entries:
            _deleted_entries[k].update(_existing_entries.pop(k))

@chrono
def integrate_key_value_querying(c : Call):
    """
    Creates a KeyEntry node and marks it as read by thea key Handle.
    """
    from ...logger import logger
    from ..graph import Graph
    if c.status == 1:
        logger.debug("Querying key entry.")
        if c.arguments.key_handle.lower() not in _inverted_handles:
            logger.warning("Trying to query value with unseen key handle.")
            return
        
        # Finding the open key handle :
        h = _inverted_handles[c.arguments.key_handle.lower()]
        k = h.key

        # Finding the last key in the path to the entry :
        if c.name == "NtEnumerateValueKey":
            sk_path, name = split_key_and_entry_names(c.arguments.regkey + "\\" + c.arguments.key_name)
        else:
            sk_path, name = split_key_and_entry_names(c.arguments.regkey)
        sk = create_key_tree(sk_path)

        skp = sk
        while skp:
            if skp is k:
                break
            skp = skp.parent_key
            if not skp:
                logger.warning("There might be an unkown symbolic link in the registry or there is a problem:\nQuerying key '{}' from open key '{}'.".format(sk.path, k.path))

        # Creating a new KeyEntry object
        nv = KeyEntry()
        nv.reg_type = c.flags.reg_type
        nv.process_value(c.arguments.value)
        nv.name = name

        # Looking for an olod value for this entry
        ov = find_last_key_entry(sk, nv.name)
        if not ov or ov.reg_type != nv.reg_type or ov.value != nv.value or (k in _deleted_entries and ov.name in _deleted_entries[k] and _deleted_entries[k][ov.name] is ov):    # New value is the first or is different : link to previous if there is one.
            v = nv
            if ov:
                ChangesTowards(ov, nv)
        else:                                                       # Old value is still good : delete new value.
            v = ov
            for g in Graph.active_graphs():
                g.remove(nv)
        
        if k not in _existing_entries:
            _existing_entries[k] = {}
        _existing_entries[k][v.name] = v
        if v not in sk.neighbors():
            HasEntry(sk, v)
        QueriesEntry(v, h)

@chrono
def integrate_key_value_setting(c : Call):
    """
    Creates a KeyEntry node and marks it as written by a key Handle.
    """
    from ...logger import logger
    from ..graph import Graph
    if c.status == 1:
        logger.debug("Setting key entry.")
        if c.arguments.key_handle.lower() not in _inverted_handles:
            logger.warning("Trying to set value with unseen key handle.")
            return
        
        # Finding the open key handle :
        h = _inverted_handles[c.arguments.key_handle.lower()]
        k = h.key

        # Finding the last key in the path to the entry :
        sk_path, name = split_key_and_entry_names(c.arguments.regkey)
        sk = create_key_tree(sk_path)

        skp = sk
        while skp:
            if skp is k:
                break
            skp = skp.parent_key
            if not skp:
                logger.warning("There might be an unkown symbolic link in the registry or there is a problem:\nSetting key '{}' from open key '{}'.".format(sk.path, k.path))

        # Creating a new KeyEntry object
        nv = KeyEntry()
        nv.reg_type = c.flags.reg_type
        nv.process_value(c.arguments.value)
        nv.name = name

        # Looking for an olod value for this entry
        ov = find_last_key_entry(sk, nv.name)
        if not ov or ov.reg_type != nv.reg_type or ov.value != nv.value or (k in _deleted_entries and ov.name in _deleted_entries[k] and _deleted_entries[k][ov.name] is ov):    # New value is the first or is different : link to previous if there is one.
            v = nv
            if ov:
                ChangesTowards(ov, nv)
        else:                                                       # Old value is still good : delete new value.
            v = ov
            for g in Graph.active_graphs():
                g.remove(nv)
        
        if k not in _existing_entries:
            _existing_entries[k] = {}
        _existing_entries[k][v.name] = v
        if v not in sk.neighbors():
            HasEntry(sk, v)
        SetsEntry(h, v)

@chrono
def integrate_key_value_deleting(c : Call):
    """
    Creates a KeyEntry node and marks it as deleted by a key Handle.
    """
    from ...logger import logger
    from ..graph import Graph
    from ..config import ColorMap
    if c.status == 1:
        logger.debug("Deleting key entry.")
        if c.arguments.key_handle.lower() not in _inverted_handles:
            logger.warning("Trying to set value with unseen key handle.")
            return
        
        # Finding the open key handle :
        h = _inverted_handles[c.arguments.key_handle.lower()]
        k = h.key

        # Finding the last key in the path to the entry :
        sk_path, name = split_key_and_entry_names(c.arguments.regkey)
        sk = create_key_tree(sk_path)

        skp = sk
        while skp:
            if skp is k:
                break
            skp = skp.parent_key
            if not skp:
                logger.warning("There might be an unkown symbolic link in the registry or there is a problem:\nDeleting key '{}' from open key '{}'.".format(sk.path, k.path))

        # Creating a new KeyEntry object
        nv = KeyEntry()
        nv.name = name

        # Looking for an old value for this entry
        ov = find_last_key_entry(sk, nv.name)
        if not ov:                                                  # New value is the first.
            v = nv
        else:                                                       # Old value : link to new value and delete new.
            v = nv
            ChangesTowards(ov, nv)
            nv.reg_type = ov.reg_type
            nv.process_value(ov.value)
            for g in Graph.active_graphs():
                g.remove(nv)
        
        if k not in _existing_entries:
            _existing_entries[k] = {}
        _existing_entries[k][v.name] = v
        if k not in _deleted_entries:
            _deleted_entries[k] = {}
        _deleted_entries[k][v.name] = v
        v.color = ColorMap.get_config().Deleted
        if v not in sk.neighbors():
            HasEntry(sk, v)
        DeletesEntry(h, v)





# Key opening
CallHandler(lambda c : c.name in {"RegOpenKeyExA", "NtOpenKeyEx", "NtOpenKey", "RegOpenKeyExW"}, integrate_key_opening)

# Key creation
CallHandler(lambda c : c.name in {"NtCreateKey", "RegCreateKeyExW", "RegCreateKeyExA"}, integrate_key_opening)

# Key enumeration
CallHandler(lambda c : c.name in {'RegEnumKeyW', 'NtEnumerateKey', 'RegEnumKeyExW', 'RegEnumKeyExA'}, integrate_key_enumeration)

# Key closing
CallHandler(lambda c : c.name in {'RegCloseKey'}, integrate_key_closing)

# Key deleting
CallHandler(lambda c : c.name in {'NtDeleteKey', 'RegDeleteKeyW', 'RegDeleteKeyA'}, integrate_key_deleting)

# Key entry querying
CallHandler(lambda c : c.name in {'NtQueryValueKey', 'RegQueryValueExW', 'RegQueryValueExA', "RegEnumValueW", 'RegEnumValueA', 'NtEnumerateValueKey'}, integrate_key_value_querying)

# Key entry setting
CallHandler(lambda c : c.name in {'RegSetValueExW', 'RegSetValueExA', 'NtSetValueKey'}, integrate_key_value_setting)

# Key entry deleting
CallHandler(lambda c : c.name in {'RegDeleteValueW', 'RegDeleteValueA', 'NtDeleteValueKey'}, integrate_key_value_deleting)





__N_referencing_phase = BuildingPhase.request_finalizing_phase()

def find_filesystem_references(ev : BuildingPhase):
    """
    When called with the right finalizing phase event, will cause all KeyEntry nodes to link to File or Folder nodes that their value reference if these entries were modified.
    """
    from ...logger import logger
    from ..utils import is_path, path_factory, parse_command_line
    from .network import Host
    from .filesystem import File, Directory, Contains, HasDrive
    if ev.major == "Finalizer" and ev.minor == __N_referencing_phase:
        logger.debug("Finding references to filesystem in {} registry key entries.".format(len(KeyEntry)))
        for ke in KeyEntry:

            if ke.reg_type not in {"REG_SZ", "REG_MULTI_SZ"}:
                continue

            ok = False

            for e in ke.edges:
                if isinstance(e, SetsEntry | DeletesEntry):
                    ok = True
                    break
            if not ok:
                continue

            ok = False
            
            if ke.reg_type == "REG_SZ":
                base_possibilities : list[str] = [ke.value]
            else:
                base_possibilities : list[str] = ke.value

            possibilities = []
            for p in base_possibilities:
                try:
                    args = parse_command_line(p)
                    possibilities.extend(args)
                except KeyboardInterrupt:
                    raise
                finally:
                    possibilities.append(p)
            
            for p in possibilities:
                if is_path(p) and p:
                    p = path_factory(p)
                    current : Host | Directory | File = Host.current
                    next : Directory | File | None
                    parts = list(p.parts)
                    if ":" in parts[0]:
                        parts[0] = parts[0].replace("\\", "")
                    for pi in parts:
                        next = None
                        for e in current.edges:
                            if isinstance(e, HasDrive | Contains) and e.source is current and e.destination.name.lower() == pi.lower():
                                next = e.destination
                                break
                        if next is None:
                            break
                        current = next
                    else:
                        ReferencesFileSystem(ke, current)
        
        logger.debug("Got {} references to filesystem points.".format(len(ReferencesFileSystem)))

BuildingPhase.add_callback(find_filesystem_references)




del logger, Optional, Union, Color, Arrow, UniqueVertex, Vertex, Call, CallHandler, Host, Process, File, Directory, Any, chrono, integrate_key_opening, integrate_key_enumeration, integrate_key_closing, integrate_key_deleting, integrate_key_value_querying, integrate_key_value_setting, integrate_key_value_deleting, find_filesystem_references


"""
This module defines an evaluator class, which is a pickable Python function.
"""

from typing import Any, Generic, TypeVar, ParamSpec




X = TypeVar("X")
R = TypeVar("R")

class Evaluator(Generic[X, R]):

    """
    Allows you to define simple Python functions that can be pickled without existing in a static module.
    To create one, you must give an evaluable expression containing a single global variable x.
    """
    
    def __init__(self, code : str) -> None:
        from typing import Callable
        self.__code = ""
        self.__compiled : Callable[[X], R] | None = None
        self.code = code

    def __compile(self):
        """
        Compiles the function with the associated evaluable code.
        """
        try:
            compile(self.__code, "test_compile", "eval")
        except Exception as E:
            raise E from None
        code = "def evaluator_func(x):\n\treturn " + self.__code
        env = {}
        exec(code, env)
        self.__compiled = env["evaluator_func"]
    
    @property
    def code(self) -> str:
        """
        The evaluable code of function (Python code).
        """
        return self.__code
    
    @code.setter
    def code(self, code : str):
        if not isinstance(code, str):
            raise TypeError("Expected str, got " + repr(type(code).__class__))
        try:
            old_code, self.__code = self.__code, code
            self.__compile()
        except SyntaxError as E:
            self.__code = old_code
            raise SyntaxError("Cannot set function code as it contains a syntax error.") from E
        self.__code = code
    
    def __str__(self) -> str:
        return "lambda x : " + self.__code
    
    def __repr__(self) -> str:
        return "Evaluator(" + repr(self.__code) + ")"
    
    def __call__(self, x : X) -> R:
        try:
            return self.__compiled(x)
        except BaseException as E:
            raise E from None
    
    def __getstate__(self) -> dict[str, Any]:
        return {"code" : self.__code}
    
    def __setstate__(self, state : dict[str, Any]):
        self.__code = ""
        self.__compiled = None
        for name, value in state.items():
            setattr(self, name, value)


from typing import Any, Callable, Generic, Iterable, Iterator, Optional, TypeVar
from ...bakery.source.graph import Arrow, Edge, Graph, UniqueVertex, Vertex
from ...bakery.source.colors import Color
from ...bakery.source.types.execution import *
from ...bakery.source.types.network import *
from ...bakery.source.types.filesystem import *



V = TypeVar("V", bound=Vertex)

class MetaVertex(UniqueVertex, Generic[V]):

    """
    This class describes a type vertex. One or multiple Vertex subclasses can be associated to it.
    """

    @classmethod
    def __class_getitem__(cls, cls_init):
        try:
            Mv = cls()
            Mv.cls = cls_init
            return Mv
        except:
            return super().__class_getitem__(cls_init)

    __slots__ = {
        "__class" : "The class that this MetaVertex represents.",
        "__condition" : "An additional condition function. Takes a Vertex as an input and tells whether or not it can be a valid match (does not need to check type)",
        "__color" : "The Color forcefully set to this MetaVertex."
    }

    __pickle_slots__ = {
        "cls",
        "condition"
    }

    def __init__(self, *, c: Color | None = None, parent: Optional[Vertex] = None) -> None:
        from ...bakery.source.colors import Color
        super().__init__(c=c, parent=parent)
        self.__class : type[Vertex] = Vertex
        self.__condition : Callable[[Vertex], bool] | bool = True
        self.__color : Color | None = c
        self.edges : set[MetaEdge]

    @property
    def color(self) -> Color:
        """
        The color of this MetaVertex. Defaults to the average of the colors of all its Vertex classes.
        """
        from ...bakery.source.colors import Color
        from ...bakery.source.config import ColorMap
        if self.__color !=  None:
            return self.__color
        if isinstance(self.cls, type):
            if not hasattr(self.cls, "color") or not isinstance(self.cls.color, Color):
                return ColorMap.get_config().get_color(self.cls)
            return self.cls.color
        else:
            return Color.average(*[(cls.color if hasattr(cls, "color") and isinstance(cls.color, Color) else ColorMap.get_config().get_color(cls)) for cls in self.cls.__args__])
    
    @color.setter
    def color(self, c : Color):
        from ...bakery.source.colors import Color
        if not isinstance(c, Color):
            raise TypeError("Expected Color, got " + repr(type(c).__name__))
        self.__color = c
    
    @property
    def cls(self) -> type[V]:
        """
        The class that his vertex represents.
        """
        return self.__class
    
    @cls.setter
    def cls(self, cls : type[V]):
        """
        Sets the class associated with this vertex.
        """
        from typing import _UnionGenericAlias
        from types import UnionType
        from .utils import class_union
        from ...bakery.source.graph import Vertex
        if isinstance(cls, type) and issubclass(cls, Vertex):
            self.__class = cls
        elif isinstance(cls, _UnionGenericAlias):
            cls = class_union(*cls.__args__)
            for c in cls.__args__:
                if not isinstance(c, type) or not issubclass(c, Vertex):
                    raise TypeError("Expected subclasses of Vertex , got " + repr(c))
            self.__class = cls
        elif isinstance(cls, UnionType):
            for c in cls.__args__:
                if not isinstance(c, type) or not issubclass(c, Vertex):
                    raise TypeError("Expected subclasses of Vertex , got " + repr(c))
            self.__class = cls
        else:
            raise TypeError("Expected subclass of Vertex , got " + repr(cls))

    @property
    def condition(self) -> Callable[[Vertex], bool]:
        """
        An additional condition to check when trying to match a Vertex to this MetaVertex.
        """
        return self.__condition
    
    @condition.setter
    def condition(self, cond : Callable[[Vertex], bool]):
        """
        Sets the additional condition function for this MetaVertex.
        """
        if not callable(cond) and not isinstance(cond, bool):
            raise TypeError("Expected callable or bool, got " + repr(type(cond).__name__))
        self.__condition = cond
    
    def match(self, v : Vertex) -> bool:
        """
        Returns True if the Vertex v has a matching type.
        """
        return isinstance(v, self.__class) and (self.__condition(v) if callable(self.__condition) else self.__condition)
    
    def __get_cls_str(self) -> str:
        """
        Returns a string to display the class of a MetaVertex.
        """
        if isinstance(self.__class, type):
            return self.__class.__name__
        else:
            return str(self.__class)

    @property
    def label(self) -> str:
        return "Vertex[" + self.__get_cls_str() + "]" + ("{" + str(self.__condition) + "}" if callable(self.__condition) else "")

    def __str__(self) -> str:
        return type(self).__name__ + "[" + self.__get_cls_str() + "]" + ("*" if callable(self.__condition) else "")
    
    def __repr__(self) -> str:
        return type(self).__name__ + "[" + self.__get_cls_str() + "]" + ("{" + str(self.__condition) + "}" if callable(self.__condition) else "")
    




E = TypeVar("E", bound=Edge)

class MetaEdge(Edge, Generic[E]):

    """
    This class describes a type edge. One or multiple Edge subclasses can be associated to it.
    Note that a MetaEdge can represent a subclass of Arrow and in this case, the match can be made in both directions.
    """

    @classmethod
    def __class_getitem__(cls, cls_init):
        try:
            Mv = cls()
            Mv.cls = cls_init
            return Mv
        except:
            return super().__class_getitem__(cls_init)

    __slots__ = {
        "__class" : "The class that this MetaEdge represents.",
        "__condition" : "An additional condition function. Takes an Edge as an input and tells whether or not it can be a valid match (does not need to check type)",
        "__color" : "The Color forcefully set to this MetaVertex."
    }

    __pickle_slots__ = {
        "cls",
        "condition"
    }

    def __init__(self, source: MetaVertex, destination: MetaVertex, *, c: Color | None = None, auto_write: bool = True) -> None:
        from ...bakery.source.colors import Color
        if not isinstance(source, MetaVertex) or not isinstance(destination, MetaVertex):
            raise TypeError("Expected two MetaVertices, got " + repr(type(source).__name__) + " and " + repr(type(destination).__name__))
        super().__init__(source, destination, c=c, auto_write=auto_write)
        self.__class : type[Edge] = Edge
        self.__condition : Callable[[Edge], bool] | bool = True
        self.__color : Color | None = c
        self.source : MetaVertex
        self.destination : MetaVertex

    @property
    def color(self) -> Color:
        """
        The color of this MetaEdge. Defaults to the average of the colors of all its Edge classes.
        """
        from ...bakery.source.colors import Color
        from ...bakery.source.config import ColorMap
        if self.__color !=  None:
            return self.__color
        if isinstance(self.cls, type):
            if not hasattr(self.cls, "color") or not isinstance(self.cls.color, Color):
                return ColorMap.get_config().get_color(self.cls)
            return self.cls.color
        else:
            return Color.average(*[(cls.color if hasattr(cls, "color") and isinstance(cls.color, Color) else ColorMap.get_config().get_color(cls)) for cls in self.cls.__args__])
    
    @color.setter
    def color(self, c : Color):
        from ...bakery.source.colors import Color
        if not isinstance(c, Color):
            raise TypeError("Expected Color, got " + repr(type(c).__name__))
        self.__color = c
    
    @property
    def cls(self) -> type[E]:
        """
        The class that his edge represents.
        """
        return self.__class
    
    @cls.setter
    def cls(self, cls : type[E]):
        """
        Sets the class associated with this edge.
        """
        from typing import _UnionGenericAlias
        from types import UnionType
        from .utils import class_union
        from ...bakery.source.graph import Edge
        if isinstance(cls, type) and issubclass(cls, Edge):
            self.__class = cls
        elif isinstance(cls, _UnionGenericAlias):
            cls = class_union(*cls.__args__)
            for c in cls.__args__:
                if not isinstance(c, type) or not issubclass(c, Edge):
                    raise TypeError("Expected subclasses of Edge , got " + repr(c))
            self.__class = cls
        elif isinstance(cls, UnionType):
            for c in cls.__args__:
                if not isinstance(c, type) or not issubclass(c, Edge):
                    raise TypeError("Expected subclasses of Edge , got " + repr(c))
            self.__class = cls
        
    @property
    def condition(self) -> Callable[[Edge], bool]:
        """
        An additional condition to check when trying to match an Edge to this MetaEdge.
        """
        return self.__condition
    
    @condition.setter
    def condition(self, cond : Callable[[Edge], bool]):
        """
        Sets the additional condition function for this MetaEdge.
        """
        if not callable(cond) and not isinstance(cond, bool):
            raise TypeError("Expected callable or bool, got " + repr(type(cond).__name__))
        self.__condition = cond
    
    def match(self, e : Edge) -> bool:
        """
        Returns True if the Egde e has a matching type.
        """
        return isinstance(e, self.__class) and (self.__condition(e) if callable(self.__condition) else self.__condition)
    
    def __get_cls_str(self) -> str:
        """
        Returns a string to display the class of a MetaEdge.
        """
        if isinstance(self.__class, type):
            return self.__class.__name__
        else:
            return str(self.__class)

    @property
    def label(self) -> str:
        return "Edge[" + self.__get_cls_str() + "]" + ("{" + str(self.__condition) + "}" if callable(self.__condition) else "")

    def __str__(self) -> str:
        return type(self).__name__ + "[" + self.__get_cls_str() + "]" + ("*" if callable(self.__condition) else "")
    
    def __repr__(self) -> str:
        return type(self).__name__ + "[" + self.__get_cls_str() + "]" + ("{" + str(self.__condition) + "}" if callable(self.__condition) else "")

    def __getitem__(self, cls : type[E]) -> "MetaEdge":
        try:
            self.__class = cls
            return self
        except BaseException as E:
            raise E from None




A = TypeVar("A", bound=Arrow)

class MetaArrow(Arrow, MetaEdge, Generic[A]):

    """
    This class describes a type arrow. One or multiple Arrow subclasses can be associated to it.
    """

    __slots__ = {
        "__class" : "The class that this MetaArrow represents.",
        "__condition" : "An additional condition function. Takes an Arrow as an input and tells whether or not it can be a valid match (does not need to check type)"
    }

    __pickle_slots__ = {
        "cls",
        "condition"
    }

    def __init__(self, source: MetaVertex, destination: MetaVertex, *, c: Color | None = None, auto_write: bool = True) -> None:
        super().__init__(source, destination, c=c, auto_write=auto_write)
        self.__class : type[Arrow] = Arrow
        self.__condition : Callable[[Arrow], bool] | bool = True
    
    @property
    def cls(self) -> type[A]:
        """
        The class that his arrow represents.
        """
        return self.__class
    
    @cls.setter
    def cls(self, cls : type[A] | Iterable[type[A]]):
        """
        Sets the class associated with this arrow.
        """
        from typing import _UnionGenericAlias
        from types import UnionType
        from .utils import class_union
        from ...bakery.source.graph import Arrow
        if isinstance(cls, type) and issubclass(cls, Arrow):
            self.__class = cls
        elif isinstance(cls, _UnionGenericAlias):
            cls = class_union(*cls.__args__)
            for c in cls.__args__:
                if not isinstance(c, type) or not issubclass(c, Arrow):
                    raise TypeError("Expected subclasses of Arrow , got " + repr(c))
            self.__class = cls
        elif isinstance(cls, UnionType):
            for c in cls.__args__:
                if not isinstance(c, type) or not issubclass(c, Arrow):
                    raise TypeError("Expected subclasses of Arrow , got " + repr(c))
            self.__class = cls

    @property
    def condition(self) -> Callable[[Arrow], bool]:
        """
        An additional condition to check when trying to match an Arrow to this MetaArrow.
        """
        return self.__condition
    
    @condition.setter
    def condition(self, cond : Callable[[Arrow], bool]):
        """
        Sets the additional condition function for this MetaArrow.
        """
        if not callable(cond) and not isinstance(cond, bool):
            raise TypeError("Expected callable or bool, got " + repr(type(cond).__name__))
        self.__condition = cond
    
    def match(self, a : Arrow) -> bool:
        """
        Returns True if the Arrow a has a matching type.
        """
        return isinstance(a, self.__class) and (self.__condition(a) if callable(self.__condition) else self.__condition)

    def __get_cls_str(self) -> str:
        """
        Returns a string to display the class of a MetaArrow.
        """
        if isinstance(self.__class, type):
            return self.__class.__name__
        else:
            return str(self.__class)

    @property
    def label(self) -> str:
        return "Arrow[" + self.__get_cls_str() + "]" + ("{" + str(self.__condition) + "}" if callable(self.__condition) else "")
    
    def __str__(self) -> str:
        return type(self).__name__ + "[" + self.__get_cls_str() + "]" + ("*" if callable(self.__condition) else "")
    
    def __repr__(self) -> str:
        return type(self).__name__ + "[" + self.__get_cls_str() + "]" + ("{" + str(self.__condition) + "}" if callable(self.__condition) else "")
    
    def __getitem__(self, cls : type[A]) -> "MetaArrow":
        try:
            self.__class = cls
            return self
        except BaseException as E:
            raise E from None
    




class MetaGraph(Graph, Generic[V, E, A]):

    """
    This particular type of graph can only contain MetaVertices, MetaEdges or MetaArrows. It can be used for normal graph exploration.
    """

    __slots__ = {
        "__named_objects" : "A dict holding named vertices and edges."
    }

    vertices : set[MetaVertex]
    edges : set[MetaEdge | MetaArrow]

    def __init__(self, g : Graph = Graph()) -> None:
        if not isinstance(g, Graph):
            raise TypeError("Exepcted Graph, got " + repr(type(g).__name__))
        super().__init__()
        self.__named_objects : dict[str, MetaEdge | MetaVertex] = {}
        translation_table : dict[Vertex | Edge, MetaVertex | MetaEdge] = {}
        for v, e, u in g.pairs():
            if v in translation_table:
                Mv = translation_table[v]
            else:
                Mv = MetaVertex()
                Mv.cls = type(v)
                translation_table[v] = Mv
                self.append(Mv)
            if u in translation_table:
                Mu = translation_table[u]
            else:
                Mu = MetaVertex()
                Mu.cls = type(u)
                translation_table[u] = Mu
                self.append(Mu)
            if e in translation_table:
                Me = translation_table[e]
            else:
                Me = MetaEdge(Mv, Mu)
                Me.cls = type(e)
                translation_table[e] = Me
                self.append(Me)
    
    @property
    def names(self) -> list[str]:
        """
        Returns the list of names for named Meta{Vertices, Edges, Arrows} available in this MetaGraph.
        """
        return list(self.__named_objects)

    def add_vertex(self, v: MetaVertex[V], explore: bool = True) -> None:
        if not isinstance(V, MetaVertex):
            raise TypeError("Expected MetaVertex, got " + repr(type(v).__name__))
        return super().add_vertex(v, explore)

    def pairs(self) -> Iterator[tuple[V, E | A, V]]:
        return super().pairs()
    
    def append(self, value: V | E | A, explore: bool = False):
        if not isinstance(value, MetaVertex | MetaEdge | MetaArrow):
            raise TypeError("Expected MetaVertex, MetaEdge or MetaArrow, got " + repr(type(value).__name__))
        return super().append(value, explore)
    
    def extend(self, values: Iterable[V | E | A], explore: bool = False):
        from typing import Iterable
        if not isinstance(explore, bool):
            raise TypeError("Expected bool for explore, got " + repr(explore.__class__.__name__))
        if not isinstance(values, Iterable):
            raise TypeError("Expected iterable, got " + repr(values.__class__.__name__))
        def checked():
            for v in values:
                if not isinstance(v, MetaVertex | MetaEdge | MetaArrow):
                    raise TypeError("Expected iterable of MetaVertex, MetaEdge or MetaArrow, got " + repr(type(v).__name__))
                yield v
        return super().extend(checked(), explore)

    def __dir__(self) -> list[str]:
        return super().__dir__() + self.names
    
    def __getattribute__(self, name: str) -> Any:
        try:
            return super().__getattribute__(name)
        except AttributeError as e:
            if name not in self.__named_objects:
                raise e from None
            return self.__named_objects[name]
    
    def __getstate__(self) -> dict[str, Any]:
        d = super().__getstate__()
        d["__named_objects"] = self.__named_objects
        return d
    
    def __setstate__(self, state: dict[str, Any]):
        names = state.pop("__named_objects")
        self.__named_objects = names
        return super().__setstate__(state)
    
    def __setattr__(self, name: str, value: Any) -> None:
        try:
            super().__getattribute__(name)
        except AttributeError:
            if isinstance(value, MetaEdge | MetaVertex):
                self.__named_objects[name] = value
                self.append(value)
            else:
                return super().__setattr__(name, value)
    
    def __delattr__(self, name: str) -> None:
        if name in self.__named_objects:
            v = self.__named_objects.pop(name)
            if v not in self.__named_objects.values():
                self.remove(v)
        else:
            return super().__delattr__(name)
    
    def remove(self, value : MetaVertex | MetaEdge):
        for name, v in self.__named_objects.copy().items():
            if v == value:
                self.__named_objects.pop(name)
        super().remove(value)
    
    def __neighborhood_mapper(self, g : Graph) -> dict[Vertex, set[MetaVertex]]:
        """
        Searches the given graph for matches of the metagraph.
        """

        mapping : dict[Vertex, set[MetaVertex]] = {}

        for v in g.vertices:
            for Mv in self.vertices:
                if Mv.match(v):
                    if v not in mapping:
                        mapping[v] = set()
                    mapping[v].add(Mv)
        
        return mapping
    
    def __clear_mapper_deadends(self, mapping : dict[Vertex, set[MetaVertex]]):
        """
        Clears the deadends in a mapping returned by self.__neighborhood_mapper():
        (Operates on the mapping itself)
        """

        work = mapping.copy()

        while work:

            v, sMv = work.popitem()

            to_check_again : set[tuple[Vertex, MetaVertex]] = set()

            for Mv in sMv.copy():

                ok = False
                edge_possibilities : dict[MetaEdge, set[Edge]] = {}     # For all MetaNeighbors. If any of them has an empty 

                for Me in Mv.edges:

                    edge_possibilities[Me] = set()

                    for e in v.edges:

                        if Me.match(e) and (not isinstance(Me, MetaArrow) or (e.source is v) == (Me.source is Mv)):     # MetaEdge or well-oriented MetaArrow
                            edge_possibilities[Me].add(e)
                                    
                all_ok = True
                decisive_neighbors : set[tuple[Vertex, MetaVertex]] = set()

                for Me in edge_possibilities:

                    ok = False

                    for e in edge_possibilities[Me]:

                        u = e.source if e.source is not v else e.destination
                        Mu = Me.source if Me.source is not Mv else Me.destination
                        decisive_neighbors.add((u, Mu))
                        
                        if u in mapping and Mu in mapping[u]:
                            ok = True
                
                    if not ok:
                        all_ok = False
                
                if not all_ok:
                    mapping[v].discard(Mv)
                    to_check_again.update(decisive_neighbors)
            
            new_work : dict[Vertex, set[MetaVertex]] = {}

            for v, Mv in to_check_again:
                sMv = work.pop(v, set())
                if v not in new_work:
                    new_work[v] = sMv
                new_work[v].add(Mv)

            work.update({v : sMv.intersection(mapping[v] if v in mapping else set()) for v, sMv in new_work.items()})
                    
        # for v, sMv in list(mapping.items()):
        #     if not sMv:
        #         mapping.pop(v)
            
    def __expand_subgraph(self, g : Graph, mapping : dict[Vertex, set[MetaVertex]], sub : dict[MetaVertex | MetaEdge, Vertex | Edge], Mv : MetaVertex) -> Iterator[dict[MetaVertex | MetaEdge, Vertex | Edge]]:
        """
        Yields all the next steps of the subgraph sub of Graph g in construction by adding all possible vertices that can fit the role of Mv.
        """
        from itertools import product

        existing : dict[Vertex, MetaVertex] = {sub[Mu] : Mu for Mu in Mv.neighbors() if Mu in sub}      # The neighbors of Mv that have already been chosen

        if not existing:        # We are starting from zero or we reached a new component
            
            used_Mvs = set(v for v in sub.values() if v in mapping and Mv in mapping[v])        # The candidates for Mv that are already part of the subgraph

            for v in mapping:
                if Mv in mapping[v] and v not in used_Mvs:
                    subv = sub.copy()
                    subv[Mv] = v
                    yield subv
            
            return

        existing_iter = iter(existing)

        u = next(existing_iter)

        vertex_possibilities : set[Vertex] = {v for v in u.neighbors() if v in mapping and Mv in mapping[v]}

        for u in existing_iter:
            vertex_possibilities.intersection_update({v for v in u.neighbors() if v in mapping and Mv in mapping[v]})
        
        vertex_possibilities.difference_update(sub.values())

        for v in vertex_possibilities:
            edge_possibilities : dict[MetaEdge, set[Edge]] = {}

            for u, Mu in existing.items():

                edges = u.edges & v.edges
                Medges = Mu.edges & Mv.edges
                for Me in Medges:
                    edge_possibilities[Me] = set()
                    for e in edges:
                        if Me.match(e) and (not isinstance(Me, MetaArrow) or (e.source is v) == (Me.source is Mv)):
                            edge_possibilities[Me].add(e)
            
            Me_list = list(edge_possibilities)
            for e_list in product(*[edge_possibilities[Me] for Me in Me_list]):
                if len(set(e_list)) == len(e_list):     # No edge was used as two different MetaEdges between v and one of its neighbors
                    subv = sub.copy()
                    subv[Mv] = v
                    for Me, e in zip(Me_list, e_list):
                        subv[Me] = e
                    yield subv

    def __discover(self) -> Iterator[MetaVertex]:
        """
        Yields successive MetaVertices by exploring the MetaGraph. The next MetaVertex yielded is either a neighbor of one of the previously yielded ones or is the first of a new connected component.
        """
        
        to_do : set[MetaVertex] = self.vertices.copy()
        done : set[MetaVertex] = set()
        component_explorable : set[MetaVertex] = set()

        while to_do:
            if not component_explorable:
                Mu = to_do.pop()
                component_explorable.add(Mu)

            else:
                component_done = True
                for Mv in component_explorable.copy():
                    sMu = set(Mv.neighbors()) - done
                    if sMu:
                        component_done = False
                        Mu = sMu.pop()
                        to_do.discard(Mu)
                        break
                    else:
                        component_explorable.remove(Mv)

                if component_done:
                    continue
                    
            component_explorable.add(Mu)
            yield Mu
            done.add(Mu)
         
    def search_iter(self, g : Graph) -> Iterator[Graph]:
        """
        Searches through g for all occurences of a subgraph that matches the metagraph.
        """
        from ...bakery.source.graph import Graph
        from typing import Iterator

        # subgraphs : list[dict[MetaVertex | MetaEdge, Vertex | Edge]] = [{}]
        # next_subgraphs : list[dict[MetaVertex | MetaEdge, Vertex | Edge]] = []

        mapping = self.__neighborhood_mapper(g)

        self.__clear_mapper_deadends(mapping)

        order = list(self.__discover())

        # print("Got {} vertices with openings.".format(len([v for v, sMv in mapping.items() if sMv])))

        def build_iter(g : Graph, mapping : dict[Vertex, set[MetaVertex]], sub : dict[MetaVertex | MetaEdge, Vertex | Edge], i : int) -> Iterator[Graph]:
            if i == len(order) - 1:     # Last MetaVertex to append
                for subi in self.__expand_subgraph(g, mapping, sub, order[i]):
                    gi = Graph()
                    gi.extend(subi.values())
                    yield gi
            else:
                for subi in self.__expand_subgraph(g, mapping, sub, order[i]):
                    yield from build_iter(g, mapping, subi, i + 1)
        
        yield from build_iter(g, mapping, {}, 0)

        # for i, Mv in enumerate(self.__discover()):
        #     for sub in subgraphs:
        #         next_subgraphs.extend(self.__expand_subgraph(g, mapping, sub, Mv))
        #     subgraphs = next_subgraphs
        #     next_subgraphs = []
        
        # for dg in subgraphs:
        #     g = Graph()
        #     g.extend(dg.values())
        #     yield g





def metapath(*cls : type[Vertex], bidirectional : bool = True) -> MetaGraph:
    
    """
    Builds a path-like MetaGraph using the given arguments. They should be a sequence of alternating Vertex subclasses and Edge/Arrow subclasses.
    If birectional is True, arrow types will be interpreted by MetaEdges and thus can be matched in both ways (a matching graph might not be an oriented path).
    Otherwise, direction is preserved.
    """

    if not isinstance(bidirectional, bool):
        raise TypeError("Expected bool for bidirectional, got " + repr(type(bidirectional).__name__))

    p = MetaGraph()
    l = list(cls)

    for i, cli in enumerate(l):
        if i % 2 == 0:  # Should be a Vertex type
            if not isinstance(cli, type) or not issubclass(cli, Vertex):
                raise TypeError("Expected alternating Vertex subclasses and Edge/Arrow subclasses, got " + repr(cli))
        else:           # Should be an Edge or Arrow
            if not isinstance(cli, type) or not issubclass(cli, Edge | Arrow):
                raise TypeError("Expected alternating Vertex subclasses and Edge/Arrow subclasses, got " + repr(cli))
    
    vertices = []
    for i, cli in enumerate(l):
        if i % 2 == 0:
            Mv = MetaVertex()
            Mv.cls = cli
            vertices.append(Mv)
    
    edges = []
    for i, cli in enumerate(l):
        if i % 2 == 1:
            if not bidirectional and issubclass(cli, Arrow):
                Ma = MetaArrow(vertices[i // 2], vertices[i // 2 + 1])
                Ma.cls = cli
                edges.append(Ma)
            else:
                Me = MetaEdge(vertices[i // 2], vertices[i // 2 + 1])
                Me.cls = cli
                edges.append(Me)
    
    p.extend(vertices)
    p.extend(edges)

    return p



# N = 0

# class Square(UniqueVertex):
#     def __init__(self, *, c: Color = Color.blue, parent: Optional["Vertex"] = None) -> None:
#         global N
#         super().__init__(c=c, parent=parent)
#         self.label = type(self).__name__[0]+str(N)
#         N += 1

# class Triangle(UniqueVertex):
#     def __init__(self, *, c: Color = Color.red, parent: Optional["Vertex"] = None) -> None:
#         global N
#         super().__init__(c=c, parent=parent)
#         self.label = type(self).__name__[0]+str(N)
#         N += 1

# class Circle(UniqueVertex):
#     def __init__(self, *, c: Color = Color.green, parent: Optional["Vertex"] = None) -> None:
#         global N
#         super().__init__(c=c, parent=parent)
#         self.label = type(self).__name__[0]+str(N)
#         N += 1

# class TriangleToTriangle(Arrow):
#     pass

# class TriangleToCircle(Arrow):
#     pass

# class CircleToSquare(Arrow):
#     pass

# class SquareToTriangle(Arrow):
#     pass



# print("Building and exporting MetaGraph.")
# MG = MetaGraph()

# T0 = MetaVertex[Triangle]
# T1 = MetaVertex[Triangle]
# C2 = MetaVertex[Circle]
# S3 = MetaVertex[Square]
# T4 = MetaVertex[Triangle]

# print(repr(T0), repr(T1), repr(C2), repr(S3), repr(T4))

# MetaArrow(T0, T1).cls = TriangleToTriangle
# MetaArrow(T1, C2).cls = TriangleToCircle
# MetaArrow(C2, S3).cls = CircleToSquare
# MetaArrow(S3, T4).cls = SquareToTriangle
# MetaArrow(T4, T1).cls = TriangleToTriangle
# MetaArrow(T1, T4).cls = TriangleToTriangle

# MG.append(T0, explore=True)

# MG.export("meta.gexf")

# print("Building Graph.")
# g = Graph()

# C0 = Circle()
# T1 = Triangle()
# S2 = Square()
# T3 = Triangle()
# T4 = Triangle()
# T5 = Triangle()
# S6 = Square()
# T7 = Triangle()
# T8 = Triangle()
# C9 = Circle()
# C10 = Circle()
# C11 = Circle()
# S12 = Square()
# T13 = Triangle()
# S14 = Square()
# C15 = Circle()
# C16 = Circle()
# S17 = Square()
# T18 = Triangle()
# T19 = Triangle()
# T20 = Triangle()
# T21 = Triangle()

# TriangleToCircle(T3, C0)
# CircleToSquare(C0, S2)
# TriangleToTriangle(T1, T3)
# TriangleToTriangle(T4, T1)
# SquareToTriangle(S2, T3)
# TriangleToTriangle(T5, T3)
# SquareToTriangle(S6, T4)
# TriangleToTriangle(T3, T8)
# TriangleToTriangle(T8, T3)
# TriangleToCircle(T3, C9)
# TriangleToCircle(T8, C9)
# TriangleToCircle(T5, C11)
# CircleToSquare(C11, S6)
# CircleToSquare(C11, S17)
# CircleToSquare(C9, S12)
# SquareToTriangle(S12, T8)
# CircleToSquare(C15, S12)
# CircleToSquare(C15, S17)
# TriangleToCircle(T7, C10)
# TriangleToTriangle(T7, T13)
# TriangleToTriangle(T13, T7)
# CircleToSquare(C10, S14)
# SquareToTriangle(S14, T13)
# CircleToSquare(C16, S12)
# CircleToSquare(C16, S14)
# SquareToTriangle(S17, T18)
# TriangleToCircle(T19, C15)
# TriangleToCircle(T20, C16)
# TriangleToCircle(T21, C16)
# SquareToTriangle(S14, T20)
# SquareToTriangle(S14, T21)
# TriangleToTriangle(T18, T19)
# TriangleToTriangle(T19, T18)
# TriangleToTriangle(T19, T20)
# TriangleToTriangle(T20, T21)
# TriangleToTriangle(T21, T20)
# TriangleToTriangle(T19, T21)

# g.append(T5, explore=True)

# print("Matching...")

# for i, subi in enumerate(MG.search_iter(g)):
#     if i == 3:
#         for v in subi.vertices:
#             v.color = color.white
#         for e in subi.edges:
#             e.color = color.white

# print("Exporting Graph.")
# g.export("graph.gexf")


"""
This module contains the extract function which is used to toast baguettes.
"""

from ..rack import BaguetteRack

__all__ = ["extract"]





def extract(rack : BaguetteRack) -> BaguetteRack:

    """
    Extracts patterns in a baguette using the data written in the given BaguetteRack object.
    Returns a BaguetteRack (the same) containing the results (and exceptions on failure).
    """

    from ..rack import BaguetteRack, TimeoutExit

    try:

        if rack.toasted:
            return rack
        
        import logging
        levels = {
            0 : logging.ERROR,
            1 : logging.WARNING,
            2 : logging.INFO,
            3 : logging.DEBUG
        }
        
        from ..bakery.logger import set_level, logger
        set_level(levels[rack.verbosity])

        logger.debug("Just change worker's verbosity level.")

        import os
        from pickle import dump, load
        from Viper.better_threading import Future, DaemonThread
        from ..bakery.source.graph import Graph

        if rack.perf:
            from ..bakery.source.utils import chrono
            chrono.enabled = True
            chrono.auto_report = True

        if not os.path.exists(rack.baguette):
            raise FileNotFoundError("Could not find baguette file: {}".format(rack.report))
        if not os.path.isfile(rack.baguette):
            raise FileExistsError("Given path to baguette file is not a file.")
        if os.path.exists(rack.extracted) and not os.path.isfile(rack.extracted):
            raise FileExistsError("Given baguette extraction file exists and is not a file.")

        result : Future[bool] = Future()

        def extract_main():
            
            try:
                logger.info("Loading baguette...")
                with rack.baguette.open("rb") as file:
                    b = load(file)

                logger.info("Loading necessary MetaGraphs...")
                patterns, pattern_names = rack.patterns, rack.pattern_names

                logger.info("Searching patterns in baguette graph...")
                matches : dict[str, list[Graph]] = {}

                for name, MG in zip(pattern_names, patterns):
                    l = list(MG.search_iter(b))
                    if l:
                        matches[name] = l

                logger.info("Exporting {} matches to file '{}'...".format(sum(len(l) for l in matches.values()), rack.extracted))
                rack.extracted.parent.mkdir(parents = True, exist_ok = True)
                with open(rack.extracted, "wb") as file:
                    dump(matches, file)

                result.set(True)
                logger.info("Done !")
            except BaseException as e:
                result.set_exception(e)

        def death_timer():
            from time import sleep
            from Viper.format import duration
            if rack.maxtime < float("inf"):
                logger.info("Death timer thread started. {} remaining.".format(duration(rack.maxtime)))
                sleep(rack.maxtime)
                logger.error("Death timer reached, about to exit.")
                result.set_exception(TimeoutExit("Toasting maxtime reached"))
            else:
                while True:
                    sleep(600)
        
        t1 = DaemonThread(target = extract_main)
        t2 = DaemonThread(target = death_timer)
        t1.start()
        t2.start()
        
        while not result.wait(0.1):
            pass
        result.result()
    
    except BaseException as e:
        from traceback import print_exc, TracebackException
        rack.exception = TracebackException.from_exception(e)
        if not isinstance(e, (KeyboardInterrupt, TimeoutExit)):
            print_exc()
    finally:
        if rack.exception is None or not issubclass(rack.exception.exc_type, (KeyboardInterrupt, TimeoutExit)):
            rack.toasted = True
        return rack





del BaguetteRack


"""
This package defines all the node types for building the graph.
"""

from os.path import dirname, basename, isfile, join
import glob
from importlib import import_module
modules = glob.glob(join(dirname(__file__), "*.py"))

__all__ = [basename(f)[:-3] for f in modules if isfile(f) and not f.endswith('__init__.py')]

l = {}
for m in __all__:
    l[m] = import_module("." + m, "baguette.bakery.source.types")
    del m
locals().update(l)

del glob, dirname, basename, isfile, join, modules, import_module, l


"""
This module can be launched to run an interactive session to work with MetaGraphs.
"""

from ...bakery.logger import set_level
import logging
set_level(logging.WARNING)
from ..metalib.utils import *
from ..metalib.evaluator import Evaluator

import_env(globals())

env = globals().copy()
env.pop("import_env")

from Viper.interactive import InteractiveInterpreter

InteractiveInterpreter(env).interact("MetaLib interactive console.\nUse save(MG, name) and load(name) to save and load MetaGraphs.\nUse entries() to get a list of all MetaGraphs available in the library.\nUse remove(name) to delete a MetaGraph from the library.\nAll useful types are loaded, including Graph and MetaGraph related types, as well as Evaluators.")


from typing import Any, Dict, Iterable, Union

_hashing = set()

class record:

    """
    A class to hold key-values informations. Similar to dict but works with attributes instead.
    """

    __slots__ = ("_record",)
    _depth = 0

    def __init__(self, **kwargs : Any) -> None:
        super().__setattr__("_record", None)

        def transform(v):
            if isinstance(v, list):
                return tuple(v)
            if isinstance(v, set):
                return frozenset(v)
            if isinstance(v, dict):
                return record(**v)
            return v

        for k, v in kwargs.items():
            try:
                setattr(self, k, v)
            except:
                pass

    def __setattr__(self, name: str, value: Any) -> None:
        # if name in super().__getattribute__("__dict__"):
        #     raise ValueError("Cannot change attribute '{}' for this object.".format(name))
        try:
            hash(value)
        except:
            raise TypeError("record object can only hold hashable types, not " + repr(value.__class__.__name__))
        # from graph import vertex, edge, graph
        # if isinstance(value, (vertex, edge, graph)):
        #     raise TypeError("Cannot hold vertex, edge or graphs in record objects")
        if super().__getattribute__("_record") == None:
            super().__setattr__("_record", {})
        super().__getattribute__("_record")[name] = value
    
    def __getattribute__(self, name: str) -> Any:
        try:
            return super().__getattribute__(name)
        except AttributeError:
            if super().__getattribute__("_record") != None and name in super().__getattribute__("_record"):
                return super().__getattribute__("_record")[name]
            else:
                raise AttributeError("No such entry : " + repr(name))
    
    def __delattr__(self, name: str) -> None:
        # if name in super().__getattribute__("__dict__"):
        #     raise ValueError("Cannot delete attribute '{}' for this object.".format(name))
        if super().__getattribute__("_record") != None and name in super().__getattribute__("_record"):
            super().__getattribute__("_record").pop(name)
            if not super().__getattribute__("_record"):
                super().__setattr__("_record", None)
    
    def __dir__(self) -> Iterable[str]:
        if super().__getattribute__("_record") == None:
            return ()
        return super().__getattribute__("_record").keys()

    def __repr__(self) -> str:
        if super().__getattribute__("_record") == None:
            return "record()"
        return "record(" + ", ".join(str(k) + " = " + repr(v) for k, v in super().__getattribute__("_record").items()) + ")"

    def __str__(self) -> str:
        if super().__getattribute__("_record") == None:
            return "record{}"
        return "record{\n\t" + "\n\t".join(str(k) + " : " + repr(v) for k, v in super().__getattribute__("_record").items()) + "\n}"
    
    def __reduce__(self) -> Union[str, tuple[Any, ...]]:
        return self.__class__, (), super().__getattribute__("_record") or {}
    
    def __copy__(self) -> "record":
        return record(**super().__getattribute__("_record"))
    
    def __deepcopy__(self, memo : Dict[int, Any]) -> "record":
        from copy import deepcopy
        return record(**{name : deepcopy(value, memo) for name, value in (super().__getattribute__("_record") or {}).items()})
    
    def __setstate__(self, state : Dict[str, Any]) -> None:
        for k, v in state.items():
            setattr(self, k, v)
    
    def __eq__(self, o: object) -> bool:
        if not isinstance(o, record):
            return False
        return super().__getattribute__("_record") == super(record, o).__getattribute__("_record")
    
    def __hash__(self) -> int:
        if super().__getattribute__("_record") == None:
            return hash(0)
        h = 0
        for k, v in (super().__getattribute__("_record") or {}).items():
            if id(k) in _hashing:
                hk = -1
            else:
                _hashing.add(id(k))
                hk = hash(k)
                _hashing.remove(id(k))
            if id(v) in _hashing:
                hv = -1
            else:
                _hashing.add(id(v))
                hv = hash(v)
                _hashing.remove(id(v))            
            h += hk - hv
        return hash(h)


del Any, Dict, Iterable, Union


"""
This module contains utilities to manage BAGUETTE data.
"""

from pathlib import Path
from traceback import TracebackException
from typing import Any, Literal, TypeVar

from .croutons.source.metagraph import MetaGraph
from .bakery.source.filters import Filter

__all__ = ["BaguetteRack", "TimeoutExit"]





class TimeoutExit(SystemExit):
    """
    This exception means that a critical timeout has been reached, causing interpreter exit when raised.
    """

T = TypeVar("T", bound="BaguetteRack")

class BaguetteRack:

    """
    Baguette racks are just a utility to organize baguettes easily.
    Remember to export() them before deleting them, as this will write the index file.
    """

    __slots__ = {
        "__working_directory" : "The absolute path to the working directory",
        "__index" : "The path to the index file. Can be relative to __working_directory",
        "__report" : "The path to the Cuckoo report file. Can be relative to __working_directory",
        "__baguette" : "The path to the baguette file. Can be relative to __working_directory",
        "__visual" : "The path to the Gephi file. Can be relative to __working_directory",
        "__extracted" : "The path to the results of the metagraph search. Can be relative to __working_directory",
        "__perf" : "A boolean indicating if a performance report should be returned",
        "__filters" : "The list if filters applied during the compilation phase",
        "__patterns" : "The list of MetaGraphs to search for during the extraction phase",
        "__skip_data_comparison" : "A boolean indicating if Data nodes were compared during compilation",
        "__skip_diff_comparison" : "A boolean indicating if Diff nodes were compared during compilation",
        "__exception" : "The last caught exception",
        "__baked" : "Indicates if the baguette has been successfully baked",
        "__toasted" : "Indicates if the baguette has been successfully toasted",
        "__verbosity" : "The verbosity level to apply when baking or toasting", 
        "__maxtime" : "The maximum amount of time to spend on this baguette"
    }

    names = {
        "working_directory" : "BAGUETTE Directory",
        "index" : "BAGUETTE Index File",
        "report" : "Cuckoo Report File",
        "baguette" : "Python BAGUETTE File",
        "visual" : "Gephi File",
        "extracted" : "Extracted MetaGraphs Python File",
    }

    def __init__(self, working_directory : str | Path | None = None) -> None:
        from pathlib import Path
        from traceback import TracebackException
        from typing import Literal
        self.__working_directory : Path | None = None
        self.__index : Path | None = None
        self.__report : Path | None = None
        self.__baguette : Path | None = None
        self.__visual : Path | None = None
        self.__extracted : Path | None = None
        self.__perf : bool = False
        self.__filters : list[str] = []
        self.__patterns : list[str] = []
        self.__skip_data_comparison : bool = False
        self.__skip_diff_comparison : bool = False
        self.__exception : TracebackException | None = None
        self.__baked : bool = False
        self.__toasted : bool = False
        self.__verbosity : Literal[0, 1, 2, 3] = 0
        self.__maxtime : float = float("inf")
        if working_directory is not None and not isinstance(working_directory, str | Path):
            raise TypeError("Expected Path, got " + repr(type(working_directory).__name__))
        if working_directory is not None:
            self.set_defaults(Path(working_directory))
        
    def set_defaults(self, working_directory : Path):
        """
        Sets the default values for parameters given the new working directory.
        """
        working_directory = working_directory.expanduser().resolve()
        self.__working_directory = working_directory
        self.__index = working_directory / "index.pyt"
        self.__baguette = working_directory / "baguette.pyt"
        self.__visual = working_directory / "visual.gexf"
        self.__extracted = working_directory / "extracted.pyt"
        self.check()

    @property
    def maxtime(self) -> float:
        """
        The maximum amount of time (in seconds) to spend on the baking or toasting process of this baguette.
        Defaults to infinity.
        """
        return self.__maxtime

    @maxtime.setter
    def maxtime(self, value : float):
        from math import isnan
        if not isinstance(value, float):
            raise TypeError("Expected float, got " + repr(type(value).__name__))
        if value <= 0 or isnan(value):
            raise ValueError("Expected positive value for timeout, got {}".format(value))
        self.__maxtime = value

    @property
    def verbosity(self) -> Literal[0, 1, 2, 3]:
        """
        The verbosity level that should be applied when baking or toasting this baguette.
        0 : Errors only (default)
        1 : Warnings
        2 : Info
        3 : Debug
        """
        return self.__verbosity

    @verbosity.setter
    def verbosity(self, value : Literal[0, 1, 2, 3]):
        if not isinstance(value, int):
            raise TypeError("Expected int, got " + repr(type(value).__name__))
        if value not in {0, 1, 2, 3}:
            raise ValueError("Verbosity should be in range(4), got {}".format(value))
        self.__verbosity = value

    @property
    def baked(self) -> bool:
        """
        Indicates if the baguette has been successfully baked.
        """
        return self.__baked
    
    @baked.setter
    def baked(self, value : bool):
        if not isinstance(value, bool):
            raise TypeError("Expected bool, got " + repr(type(value).__name__))
        self.__baked = value
    
    @property
    def toasted(self) -> bool:
        """
        Indicates if the baguette has been successfully toasted.
        """
        return self.__toasted
    
    @toasted.setter
    def toasted(self, value : bool):
        if not isinstance(value, bool):
            raise TypeError("Expected bool, got " + repr(type(value).__name__))
        self.__toasted = value
    
    @property
    def exception(self) -> TracebackException | None:
        """
        The last caught exception in the baking or toasting process if any.
        """
        return self.__exception
    
    @exception.setter
    def exception(self, value : TracebackException | None):
        from traceback import TracebackException
        if value is not None and not isinstance(value, TracebackException):
            raise TypeError("Expected TracebackException or None, got " + repr(type(value).__name__))
        self.__exception = value
        
    @property
    def perf(self) -> bool:
        """
        Indicates if the current baguette should be baked or toasted with a performance report.
        """
        return self.__perf
    
    @perf.setter
    def perf(self, value : bool):
        if not isinstance(value, bool):
            raise TypeError("Expected bool, got " + repr(type(value).__name__))
        self.__perf = value

    @property
    def patterns(self) -> list[MetaGraph]:
        """
        The MetaGraph patterns to search for during the baking of this baguette.
        """
        from .croutons import metalib
        from .croutons.source.metagraph import MetaGraph
        l = []
        for name in self.__patterns:
            p = getattr(metalib, name)
            if isinstance(p, MetaGraph):
                l.append(p)
        return l
    
    def add_pattern(self, name : str):
        """
        Adds a MetaGraph pattern from its name in the metalib.
        """
        if not isinstance(name, str):
            raise TypeError("Expected str, got " + repr(type(name).__name__))
        from .croutons import metalib
        if name not in dir(metalib):
            raise NameError("No MetaGraph named '{}'".format(name))
        self.__patterns.append(name)

    def clear_patterns(self):
        """
        Clears the MetaGraph patterns set for this baguette.
        """
        self.__patterns.clear()
    
    @property
    def pattern_names(self) -> list[str]:
        """
        The list of MetaGraphs applied to this baguette as named in the metalib.
        """
        return self.__patterns.copy()

    @pattern_names.setter
    def pattern_names(self, value : list[str]):
        if not isinstance(value, list):
            raise TypeError("Expected list, got " + repr(type(value).__name__))
        for name in value:
            if not isinstance(name, str):
                raise TypeError("Expected list of str, got a " + repr(type(name).__name__))
            
        old, self.__patterns = self.__patterns, []
        try:
            for name in value:
                self.add_pattern(name)
        except:
            self.__patterns = old
            raise

    @property
    def filters(self) -> list[Filter]:
        """
        The filters to apply/applied during the baking of this baguette.
        """
        from .bakery.source import filters
        l = []
        for name in self.__filters:
            f = getattr(filters, name)
            if isinstance(f, filters.Filter):
                l.append(f)
        return l

    def add_filter(self, name : str):
        """
        Adds a baguette filter from its name in the filters module.
        """
        if not isinstance(name, str):
            raise TypeError("Expected str, got " + repr(type(name).__name__))
        from .bakery.source import filters
        if name not in dir(filters):
            raise NameError("No Filter named '{}'".format(name))
        self.__filters.append(name)
    
    def clear_filters(self):
        """
        Clears the filters set for this baguette.
        """
        self.__filters.clear()
    
    @property
    def filter_names(self) -> list[str]:
        """
        The list of Filters applied to this baguette as named in the module filters.
        """
        return self.__filters.copy()

    @filter_names.setter
    def filter_names(self, value : list[str]):
        if not isinstance(value, list):
            raise TypeError("Expected list, got " + repr(type(value).__name__))
        for name in value:
            if not isinstance(name, str):
                raise TypeError("Expected list of str, got a " + repr(type(name).__name__))
            
        old, self.__filters = self.__filters, []
        try:
            for name in value:
                self.add_filter(name)
        except:
            self.__filters = old
            raise

    @property
    def skip_data_comparison(self) -> bool:
        """
        Indicates if Data nodes should be compared during the baking process.
        Useful for performance increase, but some information may be missing.
        """
        return self.__skip_data_comparison
    
    @skip_data_comparison.setter
    def skip_data_comparison(self, value : bool):
        if not isinstance(value, bool):
            raise TypeError("Expected bool, got " + repr(type(value).__name__))
        self.__skip_data_comparison = value

    @property
    def skip_diff_comparison(self) -> bool:
        """
        Indicates if Diff nodes should be compared during the baking process.
        Useful for performance increase, but some information may be missing.
        """
        return self.__skip_diff_comparison
    
    @skip_diff_comparison.setter
    def skip_diff_comparison(self, value : bool):
        if not isinstance(value, bool):
            raise TypeError("Expected bool, got " + repr(type(value).__name__))
        self.__skip_diff_comparison = value
    
    def check(self, index : Path | None = None, working_directory : Path | None = None, report : Path | None = None, baguette : Path | None = None, visual : Path | None = None, extracted : Path | None = None):
        """
        Checks if the current baguette is well configured.
        Raises the appropriate exception if it is not.
        """
        working_directory = self.working_directory if working_directory is None else working_directory

        index = self.index if index is None else index
        report = self.report if report is None else report
        baguette = self.baguette if baguette is None else baguette
        visual = self.visual if visual is None else visual
        extracted = self.extracted if extracted is None else extracted

        index = index if index.is_absolute() else working_directory / index
        report = report if report.is_absolute() else working_directory / report
        baguette = baguette if baguette.is_absolute() else working_directory / baguette
        visual = visual if visual.is_absolute() else working_directory / visual
        extracted = extracted if extracted.is_absolute() else working_directory / extracted

        if working_directory.exists() and not working_directory.is_dir():
            raise FileExistsError("The given working directory already exists and is not a directory.")
        if index.exists() and not index.is_file():
            raise FileExistsError("The given index file path exists and is not a file.")
        if index.parent != working_directory:
            raise ValueError("The index file must be located directly in the working directory.")
    
    @property
    def working_directory(self) -> Path:
        """
        The path that contains the index file and all other baguette files by default.
        """
        if not self.__working_directory:
            raise RuntimeError("Working directory path has not been set yet.")
        return self.__working_directory

    @working_directory.setter
    def working_directory(self, value : str | Path):
        from pathlib import Path
        if not isinstance(value, str | Path):
            raise TypeError("Expected Path, got " + repr(type(value).__name__))
        if isinstance(value, str):
            try:
                value = Path(value)
            except BaseException as e:
                raise e from None
        value = value.expanduser().resolve().absolute()
        old = self.__working_directory
        index, report, baguette, visual, extracted = None, None, None, None, None
        if old:
            if self.__index and self.__index.is_relative_to(old):
                index = value / self.__index.relative_to(old)
            if self.__report and self.__report.is_relative_to(old):
                report = value / self.__report.relative_to(old)
            if self.__baguette and self.__baguette.is_relative_to(old):
                baguette = value / self.__baguette.relative_to(old)
            if self.__visual and self.__visual.is_relative_to(old):
                visual = value / self.__visual.relative_to(old)
            if self.__extracted and self.__extracted.is_relative_to(old):
                extracted = value / self.__extracted.relative_to(old)
        self.check(working_directory = value, index = index, report = report, baguette = baguette, visual = visual, extracted = extracted)
        if old and old.is_dir():
            from os import renames
            renames(old, value)
        self.__working_directory = value
        if index:
            self.index = index
        if report:
            self.report = report
        if baguette:
            self.baguette = baguette
        if visual:
            self.visual = visual
        if extracted:
            self.extracted = extracted

    @property
    def index(self) -> Path:
        """
        The path to the index file which stores the paths to all parts of the baguette.
        """
        if not self.__index:
            return self.working_directory / "index.json"
        return self.__index
    
    @index.setter
    def index(self, value : str | Path):
        from pathlib import Path
        if not isinstance(value, str | Path):
            raise TypeError("Expected Path, got " + repr(type(value).__name__))
        if isinstance(value, str):
            try:
                value = Path(value)
            except BaseException as e:
                raise e from None
        value = value.expanduser().resolve().absolute()
        if not self.__working_directory:
            self.working_directory = value.parent
        self.check(index = value)
        if self.__index and self.__index.is_file():
            from os import renames
            renames(self.__index, value)
        self.__index = value
    
    @property
    def report(self) -> Path:
        """
        The path to the Cuckoo report of this baguette (.json).
        """
        if self.__report is None:
            self.__report = self.working_directory / "report.json"
        return self.__report

    @report.setter
    def report(self, value : str | Path):
        from pathlib import Path
        if not isinstance(value, str | Path):
            raise TypeError("Expected Path, got " + repr(type(value).__name__))
        if isinstance(value, str):
            try:
                value = Path(value)
            except BaseException as e:
                raise e from None
        value = value.expanduser().resolve().absolute()
        self.check(report = value)
        if self.__report and self.__report.is_file():
            from os import renames
            renames(self.__report, value)
        self.__report = value

    @property
    def baguette(self) -> Path:
        """
        The path to the baguette file itself (the pickle of the baguette graph) (.pyt).
        """
        if self.__baguette is None:
            self.__baguette = self.working_directory / "baguette.pyt"
        return self.__baguette

    @baguette.setter
    def baguette(self, value : str | Path):
        from pathlib import Path
        if not isinstance(value, str | Path):
            raise TypeError("Expected Path, got " + repr(type(value).__name__))
        if isinstance(value, str):
            try:
                value = Path(value)
            except BaseException as e:
                raise e from None
        value = value.expanduser().resolve().absolute()
        self.check(baguette = value)
        if self.__baguette and self.__baguette.is_file():
            from os import renames
            renames(self.__baguette, value)
        self.__baguette = value

    @property
    def visual(self) -> Path:
        """
        The path to the visual representation file of the baguette (.gexf).
        """
        if self.__visual is None:
            self.__visual = self.working_directory / "visual.gexf"
        return self.__visual

    @visual.setter
    def visual(self, value : str | Path):
        from pathlib import Path
        if not isinstance(value, str | Path):
            raise TypeError("Expected Path, got " + repr(type(value).__name__))
        if isinstance(value, str):
            try:
                value = Path(value)
            except BaseException as e:
                raise e from None
        value = value.expanduser().resolve().absolute()
        self.check(visual = value)
        if self.__visual and self.__visual.is_file():
            from os import renames
            renames(self.__visual, value)
        self.__visual = value
    
    @property
    def extracted(self) -> Path:
        """
        The path to the metagraph search results file for this baguette (.pyt).
        """
        if self.__extracted is None:
            self.__extracted = self.working_directory / "extracted.pyt"
        return self.__extracted

    @extracted.setter
    def extracted(self, value : str | Path):
        from pathlib import Path
        if not isinstance(value, str | Path):
            raise TypeError("Expected Path, got " + repr(type(value).__name__))
        if isinstance(value, str):
            try:
                value = Path(value)
            except BaseException as e:
                raise e from None
        value = value.expanduser().resolve().absolute()
        self.check(extracted = value)
        if self.__extracted and self.__extracted.is_file():
            from os import renames
            renames(self.__extracted, value)
        self.__extracted = value
    
    def export(self):
        """
        Writes the information of this baguette to the index file.
        Does nothing if the path to the index file has not been set yet.
        """
        from pickle import dump
        if not self.__working_directory:
            return
        self.index.parent.mkdir(parents = True, exist_ok = True)
        with open(self.index, "wb") as f:
            dump(self, f)
    
    @classmethod
    def import_from(cls : type[T], path : str | Path) -> T:
        """
        Loads the information on a baguette from the given index path.
        """
        from pathlib import Path
        from pickle import load
        if not isinstance(path, str | Path):
            raise TypeError("Expected Path, got " + repr(type(path).__name__))
        if isinstance(path, str):
            try:
                path = Path(path)
            except BaseException as e:
                raise e from None
        path = path.expanduser().resolve()
        with path.open("rb") as f:
            return load(f)
    




del Path, T, TypeVar, Any, TracebackException


"""
This module defines the Vertices and Edges that focus on data interpretation.
"""

from ..config import CompilationParameters
from ...logger import logger
from abc import ABCMeta
from typing import Iterator, Optional
from ..colors import Color
from ..graph import Arrow, Edge, UniqueVertex, Vertex
from ..utils import chrono
from .filesystem import File, Handle
from .network import Socket, Connection
from ..build import BuildingPhase
from typing import Sequence, Hashable, TypeVar

__all__ = ["levenshtein_similarity", "levenshtein_subset_similarity", "entropy", "Data", "IsSimilarTo", "IsAlmostIn", "register_read_operation", "register_write_operation"]





logger.debug("Loading data library.")

S = TypeVar("S", bound=Sequence[Hashable])

@chrono
def levenshtein_similarity(s1 : S, s2 : S, threshold : float = 1.0) -> float:
    """
    Returns the normalized Levenshtein distance (float between 0 and 1) between two strings, where the weight of insertion and deletion operation is the same than the one of substitution.
    If threshold is given, it must be a float representing the minimum value of similarity required to push the calculation to the end. If similarity gets lower, 0 is returned.
    """
    from typing import Sequence
    if not isinstance(s1, Sequence):
        raise TypeError("Expected two Sequences of Hashables, got " + repr(type(s1).__name__) + " and " + repr(type(s2).__name__))
    if not isinstance(s2, type(s1)):
        raise TypeError("Expected two {}s, got {} and {}".format(type(s1).__name__, repr(type(s1).__name__), repr(type(s2).__name__)))
    if not isinstance(threshold, float):
        try:
            threshold = float(threshold)
        except:
            raise TypeError("Expected float for threshold, got " + repr(type(threshold).__name__)) from None
    if not (0 <= threshold <= 1):
        raise ValueError("Threshold must be between zero and one, got " + repr(threshold))
    if not s1 and not s2:
        return 0.0
    if hash(s1) == hash(s2) and s1 == s2:
        return 1.0
    if threshold >= 1.0:
        score_threshold = None
    else:
        score_threshold = round((1 - threshold) * max(len(s1), len(s2)))
    modifier = 1
    from Levenshtein import distance
    s = 1.0 - (distance(s1, s2, weights = (1, 1, modifier), score_cutoff = score_threshold) / max(len(s1), len(s2)))
    if s <= threshold:
        return 0.0
    return s

@chrono
def levenshtein_subset_similarity(s1 : S, s2 : S, threshold : float = 1.0) -> float:
    """
    Returns the normalized Levenshtein distance (float between 0 and 1) between two strings, where the weight of insertion and deletion operation is lower than the one of substitution.
    If threshold is given, it must be a float representing the minimum value of similarity required to push the calculation to the end. If similarity gets lower, 0 is returned.
    """
    from typing import Sequence
    if not isinstance(s1, Sequence):
        raise TypeError("Expected two Sequences of Hashables, got " + repr(type(s1).__name__) + " and " + repr(type(s2).__name__))
    if not isinstance(s2, type(s1)):
        raise TypeError("Expected two {}s, got {} and {}".format(type(s1).__name__, repr(type(s1).__name__), repr(type(s2).__name__)))
    if not isinstance(threshold, float):
        try:
            threshold = float(threshold)
        except:
            raise TypeError("Expected float for threshold, got " + repr(type(threshold).__name__)) from None
    if not (0 <= threshold <= 1):
        raise ValueError("Threshold must be between zero and one, got " + repr(threshold))
    if not s1 and not s2:
        return 0.0
    if hash(s1) == hash(s2) and s1 == s2:
        return 1.0
    if threshold >= 1.0:
        score_threshold = None
    else:
        score_threshold = round((1 - threshold) * max(len(s1), len(s2)))
    modifier = 3
    from Levenshtein import distance
    s = 1.0 - (distance(s1, s2, weights = (1, 1, modifier), score_cutoff = score_threshold) / max(len(s1), len(s2)))
    if s <= threshold:
        return 0.0
    return s

@chrono
def entropy(s : str | bytes | bytearray) -> float:
    """
    Computes the character entropy of the given string.
    """
    if not s:
        return 0.0
    from math import log2
    from collections import Counter
    probabilities = {c : nc / len(s) for c, nc in Counter(s).items()}
    return sum(-p * log2(p) for p in probabilities.values())

@chrono
def printable_rate(s : bytes | bytearray) -> tuple[float, str]:
    """
    Computes the ratio of printable bytes (which can be decoded to the best endoding). Also returns the best probale encoding.
    """
    if not s:
        return 1.0, "raw"

    import codecs
    n_err = 0

    def count(err : UnicodeError) -> tuple[str, int]:
        """
        Adds one to the decoding errors counter and resume normal decoding.
        """
        nonlocal n_err
        n_err += 1
        if isinstance(err, UnicodeDecodeError | UnicodeTranslateError | UnicodeEncodeError):
            return "", err.start + 1
        else:
            raise ValueError("Unknown UnicodeError")
    
    codecs.register_error("count", count)

    best = float("inf")
    best_enc = ""

    # ("utf-8", "ascii", "utf-16", "utf-32", "gbk")

    for enc in ("ascii", ):
        n_err = 0
        codecs.decode(s, enc, "count")
        if n_err < best:
            best = n_err
            best_enc = enc
    
    return (len(s) - best) / len(s), best_enc if best == 0 else "raw"





class Data(UniqueVertex):
    
    """
    A data vertex. Represents data read, written or appended to a file.
    """

    similarity_threshold = 0.75

    __slots__ = {
        "data" : "The bytes data exchanged through the handle",
        "length" : "The length of the data in bytes",
        "entropy" : "The byte-wise entropy of the message",
        "time" : "The time at which this message was seen",
        "isprintable" : "Indicates if a message can be converted to a UTF-8 str and contains only printable characters",
        "__initialized" : "Indicates if this Vertex can be compared to other Data Vertices"
    }

    __pickle_slots__ = {
        "data",
        "length",
        "entropy",
        "time",
        "isprintable"
    }

    def __init__(self, *, c: Color | None = None, parent: Optional[Vertex] = None) -> None:
        super().__init__(c=c, parent=parent)
        self.__initialized = False
        self.data : bytes = b""
        self.length : int = 0
        self.entropy : float = 0.0
        self.time : float = 0.0
        self.isprintable : bool = False
    
    @property
    def vector(self) -> Handle | Socket:
        """
        The Handle or Socket Vertex that this Data node is Conveying data to or from.
        """
        from .filesystem import Handle
        from .network import Socket
        for u in self.neighbors():
            if isinstance(u, Handle | Socket):
                return u
        raise RuntimeError("Got a data node without vector.")
    
    @chrono
    def set_data(self, data : str):
        """
        Writes the corresponding data to the node.
        Also computes all related metrics.
        """
        self.data = data.encode("utf-8")
        self.length = len(self.data)
        self.entropy = entropy(self.data)
        self.isprintable = data.isprintable()
    
    def compare(self):
        """
        Links the node to other similar nodes.
        """
        self.__initialized = True
        for d in Data:
            if (1 - Data.similarity_threshold) * max(len(d.data), len(self.data)) < abs(len(d.data) - len(self.data)):      # Size difference is too high, similarity will be below threshold
                continue
            if d is not self and d.__initialized:
                s1 = levenshtein_similarity(d.data, self.data, Data.similarity_threshold)
                if s1 < Data.similarity_threshold:
                    continue
                if self.time < d.time:
                    l1 = IsSimilarTo(self, d)
                else:
                    l1 = IsSimilarTo(d, self)
                l1.weight = s1
                # s2 = levenshtein_subset_similarity(d.data, self.data)
                # if self.time < d.time:
                #     l2 = IsAlmostIn(self, d)
                # else:
                #     l2 = IsAlmostIn(d, self)
                # l2.weight = s2

__N_data_comparison_phase = BuildingPhase.request_finalizing_phase()

@chrono
def finalize_data_nodes(e : BuildingPhase):
    """
    When called with the right finalizing phase event, will cause all Data nodes to link themselves to similar Data nodes.
    """
    from ...logger import logger
    from ..config import CompilationParameters
    from time import time_ns
    from Viper.format import duration
    if e.major == "Finalizer" and e.minor == __N_data_comparison_phase:
        if CompilationParameters.get_config().SkipLevenshteinForDataNodes:
            return
        logger.debug("Finalizing {} Data nodes.".format(len(Data)))
        d : Data
        n = len(Data)
        t0 = time_ns()
        t = t0
        for i, d in enumerate(Data):
            d.compare()
            if (time_ns() - t) / 1000000000 > 15:
                t = time_ns()
                logger.debug("Finalizing {} Data nodes : {:.2f}%. ETA : {}".format(len(Data), (i + 1) / n * 100, duration(round((t - t0) / (i + 1) * (n - i - 1)))))

BuildingPhase.add_callback(finalize_data_nodes)





class IsSimilarTo(Arrow):

    """
    This kind of arrow indicates two Data nodes have a similar content. The direction of the arrow indicates the order of apparition.
    """

    label : str = ""

    source : Data
    destination : Data





class IsAlmostIn(Arrow):

    """
    This kind of arrow indicates that a Data node's content might be contained in another's one. The direction of the arrow indicates the order of apparition.
    """

    label : str = ""

    source : Data
    destination : Data





class DiffFile:

    """
    These objects are abstract/virtual files in which you can perform IO operations at any place.
    """

    def __init__(self) -> None:
        self.__offsets : list[int] = list()
        self.__arrays : list[bytearray] = list()
        self.__last_dump : bytes | None = None
        self.__computation_index : int = 0
        self.__last_computation_index : int = -1
    
    def arrays(self) -> Iterator[tuple[int, bytes]]:
        """
        Yields all the successives entries in the DiffFile in the form (offset, buffer).
        """
        yield from zip(self.__offsets, self.__arrays)

    @chrono
    def write(self, data : bytes, offset : int):
        """
        Writes a new block of data in the virtual diff file.
        """
        if not isinstance(data, bytes | bytearray | memoryview) or not isinstance(offset, int):
            raise TypeError("Expected readable buffer and int, got " + repr(type(data).__name__) + " and " + repr(type(offset).__name__))
        # First, insert the new block in the right place
        data = bytearray(data)
        for i, (o, a) in enumerate(self.arrays()):
            if offset <= o:
                break
        else:
            i = len(self.__offsets)
        self.__offsets.insert(i, offset)
        self.__arrays.insert(i, data)
        # Second, operate all the possible merges
        end = offset + len(data)
        old_data = bytearray(len(data))
        j = i + 1
        while j < len(self.__offsets):
            o, a = self.__offsets[j], self.__arrays[j]
            if o > end:
                break
            old_data[o - offset : o - offset + len(a)] = a
            j += 1
        old_data[:len(data)] = data
        self.__arrays[i] = old_data
        # Delete useless upcoming blocks
        for k in range(i + 1, j):
            self.__offsets.pop(i + 1)
            self.__arrays.pop(i + 1)
        # Eventually merge with previous block
        if i > 0 and self.__offsets[i - 1] + len(self.__arrays[i - 1]) >= offset:
            self.__offsets.pop(i)
            self.__arrays.pop(i)
            o, a = self.__offsets[i - 1], self.__arrays[i - 1]
            a[offset - o : offset - o + len(old_data)] = old_data
        self.__computation_index += 1
    
    @chrono
    def dump(self) -> bytes:
        """
        Returns a dump of the file content. Any empty space will be represented by '[N bytes]'.
        """
        if self.__last_computation_index != self.__computation_index:
            last_end = self.__offsets[0] if self.__offsets else 0
            d = bytearray()
            for o, a in self.arrays():
                if last_end < o:
                    d.extend("[{} bytes]".format(o - (last_end + len(a))).encode("utf-8"))
                d.extend(a)
                last_end = o + len(a)
            self.__last_computation_index = self.__computation_index
            self.__last_dump = bytes(d)
            return self.__last_dump
        else:
            if self.__last_dump is None:
                raise RuntimeError("Got an empty DiffFile dump, whereas it was marked as updated")
            return self.__last_dump





class IOOperation(metaclass = ABCMeta):

    """
    This class represents a Diff vertex atomic operation in its operation log.
    """

    def __init__(self, offset : int, data : bytes) -> None:
        self.__offset = offset
        self.__data = data
    
    @property
    def start(self) -> int:
        """
        The place in the file where this buffer starts.
        """
        return self.__offset
    
    @property
    def stop(self) -> int:
        """
        The place in the file where this buffer ends.
        """
        return self.__offset + len(self.__data)
    
    @property
    def data(self) -> bytes:
        """
        The content of the buffer.
        """
        return self.__data

    



class Read(IOOperation):

    """
    This is a single read operation in a Diff Vertex log.
    """





class Write(IOOperation):

    """
    This is a single write operation in a Diff Vertex log.
    """

    



class Diff(UniqueVertex):

    """
    A diff vertex. This represents all the information gathered on a file's content during the lifetime of a handle.
    """

    similarity_threshold = 0.90

    __slots__ = {
        "read" : "The content of the read diff file",
        "read_type" : "The file type determined by libmagic from what has been read",
        "written" : "The content of the write diff file",
        "written_type" : "The file type determined by libmagic from what has been written",
        "glob" : "The content of the global diff file",
        "glob_type" : "The file type determined by libmagic from the final state of the file",
        "read_total" : "The total amount of bytes that were read",  # Counts double if you read twice the same byte
        "read_space" : "The amount of bytes that were read in the file",    # This one only counts one
        "written_total" : "The total amount of bytes that were written",
        "written_space" : "The amount of bytes that were written in the file",
        "glob_space" : "The amount of bytes that were accessed in the file",
        "read_entropy" : "The amount of entropy that was read from the file",
        "written_entropy" : "The amount of entropy that was written to the file",
        "glob_entropy" : "The entropy that resulted in the file from all operations",
        "printable_rate" : "Indicates how much of the final state of the file only contains printable characters",
        "encoding" : "A valid encoding for the final file",
        "__initialized" : "Indicates if this Vertex can be compared to other Diff Vertices",
        "__reader_difffile" : "The DiffFile that will store the result of all reading operations",
        "__writer_difffile" : "The DiffFile that will store the result of all writing operations",
        "__glob_difffile" : "The DiffFile that will store the result of all IO operations",
        "__operations" : "The list of all operations that appear in the Diff node"
    }

    __pickle_slots__ = {
        "read",
        "read_type",
        "written",
        "written_type",
        "glob",
        "glob_type",
        "read_total",
        "read_space",
        "written_total",
        "written_space",
        "glob_space",
        "read_entropy",
        "written_entropy",
        "glob_entropy",
        "printable_rate",
        "encoding",
    }

    _active_target_diff : dict[File | Connection, "Diff"] = {}
    _active_vector_diff : dict[Handle | Socket, "Diff"] = {}

    def __init__(self, *, c: Color | None = None, parent: Optional["Vertex"] = None) -> None:
        super().__init__(c=c, parent=parent)
        self.__initialized : bool = False
        self.__reader_difffile = DiffFile()
        self.__writer_difffile = DiffFile()
        self.__glob_difffile = DiffFile()
        self.__operations : list[IOOperation] = []
        self.read : bytes = b""
        self.read_type : list[str] = []
        self.written : bytes = b""
        self.written_type : list[str] = []
        self.glob : bytes = b""
        self.glob_type : list[str] = []
        self.read_total : int = 0
        self.read_space : int = 0
        self.written_total : int = 0
        self.written_space : int = 0
        self.glob_space : int = 0
        self.read_entropy : float = 0.0
        self.written_entropy : float = 0.0
        self.glob_entropy : float = 0.0
        self.printable_rate : float = 0.0
        self.encoding : str = "raw"

    @property
    def read_buffer(self) -> bytes:
        """
        The current content of the read diff file.
        """
        return self.__reader_difffile.dump()
    
    @property
    def write_buffer(self) -> bytes:
        """
        The current content of the write diff file.
        """
        return self.__writer_difffile.dump()
    
    @property
    def glob_buffer(self) -> bytes:
        """
        The current content of the global diff file.
        """
        return self.__glob_difffile.dump()
    
    @property
    def content(self) -> tuple[bytes, bytes, bytes]:
        """
        The current content of the read, write and global diff files.
        """
        return self.read_buffer, self.write_buffer, self.glob_buffer
        
    def last_pos(self) -> int:
        """
        Returns the offset after the las operation performed in this Diff node.
        """
        if self.__operations:
            return self.__operations[-1].stop
        else:
            return 0
    
    @chrono
    def add_operation(self, op : IOOperation):
        """
        Registers a new IO operation to this Diff node.
        """
        self.__operations.append(op)
        self.__glob_difffile.write(op.data, op.start)
        if isinstance(op, Read):
            self.__reader_difffile.write(op.data, op.start)
        elif isinstance(op, Write):
            self.__writer_difffile.write(op.data, op.start)
    
    @chrono
    def compute_data(self):
        """
        Computes the values of all the attributes given the state of operations and DiffFiles.
        """
        from magic import Magic
        from ...logger import logger
        from ..config import CompilationParameters
        from ..graph import Graph

        analyzer = Magic(keep_going=True, uncompress=True)
        r, w, g = self.content
        self.read, self.written, self.glob = r, w, g
        self.read_type = analyzer.from_buffer(r).split("\\012- ")
        self.written_type = analyzer.from_buffer(w).split("\\012- ")
        self.glob_type = analyzer.from_buffer(g).split("\\012- ")
        self.read_total = sum(len(op.data) for op in self.__operations if isinstance(op, Read))
        self.written_total = sum(len(op.data) for op in self.__operations if isinstance(op, Write))
        self.read_space = len(r)
        self.written_space = len(w)
        self.glob_space = len(g)
        self.read_entropy = entropy(r)
        self.written_entropy = entropy(w)
        self.glob_entropy = entropy(g)
        self.printable_rate, self.encoding = printable_rate(g)

        if not r and w:
            for e in self.edges:
                if isinstance(e, IsDiffOf):
                    WritesInto(e.source, e.destination)
                    e.delete()
                    for G in Graph.active_graphs():
                        G.remove(e)
        elif not w and r:
            for e in self.edges:
                if isinstance(e, IsDiffOf):
                    IsReadBy(e.destination, e.source)
                    e.delete()
                    for G in Graph.active_graphs():
                        G.remove(e)

        if CompilationParameters.get_config().SkipLevenshteinForDiffNodes:
            return
        
        for u in Diff:
            if u not in self.neighbors() and u is not self:
                lw, ls, ld = 0.0, "", ""
                for sn, sb in zip(("read_buffer", "write_buffer", "global_buffer"), (r, w, g)):
                    if sb:
                        for un, ub in zip(("read_buffer", "write_buffer", "global_buffer"), u.content):
                            if (1 - Diff.similarity_threshold) * max(len(sb), len(ub)) < abs(len(sb) - len(ub)):      # Size difference is too high, similarity will be below threshold
                                continue
                            if ub:
                                s = levenshtein_similarity(sb, ub, Diff.similarity_threshold)
                                if s >= lw:
                                    lw = s
                                    ls = sn
                                    ld= un
                if ls and lw >= Diff.similarity_threshold:      # Heuristic is not perfect : checking that the threshold has indeed been reached!
                    l = HasSimilarContent(self, u)
                    l.weight = lw
                    l.source_buffer = ls
                    l.destination_buffer = ld
    
    @property
    def label(self) -> str:
        """
        The name of the node to display. It is the global data type.
        """
        if len(self.glob_type) == 1:
            return self.glob_type[0]
        return " | ".join(t for t in self.glob_type if t != "data")

    @property
    def vectors(self) -> Iterator[Handle | Socket]:
        """
        The Handle or Socket Vertices that this Diff node is interacting with.
        """
        from .filesystem import Handle
        from .network import Socket
        for u in self.neighbors():
            if isinstance(u, Handle | Socket):
                yield u
    
    @property
    def targets(self) -> Iterator[File | Connection]:
        """
        The File or Connection Vertices that this Diff node is interacting with.
        """
        from .filesystem import File
        from .network import Connection
        for u in self.neighbors():
            if isinstance(u, File | Connection):
                yield u





class IsDiffOf(Edge):

    """
    This kind of edge indicates that a Diff node sums up operations performed on the destination node.
    """

    label : str = ""

    source : File | Connection | Handle | Socket
    destination : Diff





class IsReadBy(Arrow):

    """
    This kind of arrow indicates that a Diff node only read from a vector.
    """

    source : Diff
    destination : File | Connection | Handle | Socket





class WritesInto(Arrow):

    """
    This kind of arrow indicates that a Diff node only wrote to a vector.
    """

    source : File | Connection | Handle | Socket
    destination : Diff





class HasSimilarContent(Edge):

    """
    This kind of arrow indicates that two Diff nodes have buffers with a certain similarity rate.
    """

    __slots__ = {
        "__source_buffer" : "The name of the buffer of the source node that the similarity was computed with.",
        "__destination_buffer" : "The name of the buffer of the destination node that the similarity was computed with."
    }

    __pickle_slots__ = {
        "source_buffer",
        "destination_buffer"
    }

    label : str = ""

    source : Diff
    destination : Diff

    def __init__(self, source: Vertex, destination: Vertex, *, c: Color | None = None, auto_write: bool = True) -> None:
        super().__init__(source, destination, c=c, auto_write=auto_write)
        self.__source_buffer = "global_buffer"
        self.__destination_buffer = "global_buffer"
        
    @property
    def source_buffer(self) -> str:
        """
        Returns the name of the content selected for comparison in the source node.
        """
        return self.__source_buffer
    
    @source_buffer.setter
    def source_buffer(self, name : str):
        if name not in ("read_buffer", "write_buffer", "global_buffer"):
            raise ValueError("Diff node buffers can only be set to one of ('read_buffer', 'write_buffer', 'global_buffer')")
        self.__source_buffer = name
    
    @property
    def destination_buffer(self) -> str:
        """
        Returns the name of the content selected for comparison in the destination node.
        """
        return self.__destination_buffer
    
    @destination_buffer.setter
    def destination_buffer(self, name : str):
        if name not in ("read_buffer", "write_buffer", "global_buffer"):
            raise ValueError("Diff node buffers can only be set to one of ('read_buffer', 'write_buffer', 'global_buffer')")
        self.__destination_buffer = name





@chrono
def register_read_operation(target : File | Connection, vector : Handle | Socket, content : bytes | bytearray | memoryview | str, offset : int | None = None):
    """
    Registers a read operation to be added to the corresponding DiffNodes.
    """
    if isinstance(content, str):
        content = content.encode("utf-8")
    if target not in Diff._active_target_diff:
        t = Diff()
        IsDiffOf(t, target)
        Diff._active_target_diff[target] = t
    else:
        t = Diff._active_target_diff[target]
    if vector not in Diff._active_vector_diff:
        v = Diff()
        IsDiffOf(v, vector)
        Diff._active_vector_diff[vector] = v
    else:
        v = Diff._active_vector_diff[vector]
    
    if offset == None:
        t.add_operation(Read(t.last_pos(), content))
        v.add_operation(Read(v.last_pos(), content))
    else:
        t.add_operation(Read(offset, content))
        v.add_operation(Read(offset, content))

@chrono
def register_write_operation(target : File | Connection, vector : Handle | Socket, content : bytes | bytearray | memoryview | str, offset : int | None = None):
    """
    Registers a write operation to be added to the corresponding DiffNodes.
    """
    if isinstance(content, str):
        content = content.encode("utf-8")
    if target not in Diff._active_target_diff:
        t = Diff()
        IsDiffOf(t, target)
        Diff._active_target_diff[target] = t
    else:
        t = Diff._active_target_diff[target]
    if vector not in Diff._active_vector_diff:
        v = Diff()
        IsDiffOf(v, vector)
        Diff._active_vector_diff[vector] = v
    else:
        v = Diff._active_vector_diff[vector]
    
    if offset == None:
        t.add_operation(Write(t.last_pos(), content))
        v.add_operation(Write(v.last_pos(), content))
    else:
        t.add_operation(Write(offset, content))
        v.add_operation(Write(offset, content))

__N_diff_comparison_phase = BuildingPhase.request_finalizing_phase()
__N_diff_normalization_phase = BuildingPhase.request_finalizing_phase()
__N_diff_fusion_phase = BuildingPhase.request_finalizing_phase()
        

@chrono
def finalize_diff_nodes(e : BuildingPhase):
    """
    When called with the right finalizing phase event, will cause all Diff nodes to compute their data attributes.
    """
    from ...logger import logger
    from time import time_ns
    from ..config import CompilationParameters
    from Viper.format import duration
    if e.major == "Finalizer" and e.minor == __N_diff_comparison_phase:
        logger.debug("Finalizing{} {} Diff nodes.".format(" and comparing" if not CompilationParameters.get_config().SkipLevenshteinForDiffNodes else "",len(Diff)))
        n = len(Diff)
        t0 = time_ns()
        t = t0
        for i, d in enumerate(Diff):
            d.compute_data()
            if (time_ns() - t) / 1000000000 > 15:
                t = time_ns()
                logger.debug("Finalizing {} Diff nodes : {:.2f}%. ETA : {}".format(len(Data), (i + 1) / n * 100, duration(round((t - t0) / (i + 1) * (n - i - 1)))))

@chrono
def normalize_diff_nodes(e : BuildingPhase):
    """
    When called with the right finalizing phase event, will cause all Diff nodes to compare their diff file sizes and change their size accordingly.
    """
    from typing import Type
    from ..graph import Vertex
    from ...logger import logger
    from .filesystem import File, Handle
    from .network import Socket, Connection
    from ..config import ColorMap, SizeMap
    from ..colors import Color

    def project_range(x : float, sa : float, sb : float, da : float, db : float) -> float:
        if sa == sb:
            return (da + db) / 2
        p = (x - sa) / (sb - sa)
        return p * (db - da) + da

    def normalize_neighbor(cls : Type[Vertex]):
        if len(cls) > 0:
            logger.debug("Normalizing {} {} nodes.".format(len(cls), cls.__name__))
            a, b = min(sum(v.size for v in u.neighbors() if isinstance(v, Diff)) for u in cls), max(sum(v.size for v in u.neighbors() if isinstance(v, Diff)) for u in cls)
            minsize = SizeMap.get_config().get_size(cls) * 0.75
            maxsize = SizeMap.get_config().get_size(cls) * 1.25
            for u in cls:
                u.size = project_range(sum(v.size for v in u.neighbors() if isinstance(v, Diff)), a, b, minsize, maxsize)

    if e.major == "Finalizer" and e.minor == __N_diff_normalization_phase:
        logger.debug("Normalizing sizes of {} Diff nodes.".format(len(Diff)))
        if len(Diff) == 0:
            return
        a, b = min(len(d.glob) for d in Diff), max(len(d.glob) for d in Diff)
        minsize = 0.5
        maxsize = 3.0
        for d in Diff:
            d.size = project_range(len(d.glob), a, b, minsize, maxsize)
    
        for cls in (File, Handle, Socket, Connection):
            normalize_neighbor(cls)
        
        logger.debug("Normalizing colors of {} Diff nodes.".format(len(Diff)))
        a, b = min(d.glob_entropy for d in Diff), max(d.glob_entropy for d in Diff)
        for d in Diff:
            d.color = Color.linear((ColorMap.get_config().data.DiffLowEntropy, ColorMap.get_config().data.DiffHighEntropy), (1 - d.glob_entropy / 8, d.glob_entropy / 8))

@chrono
def fuse_diff_nodes(e : BuildingPhase):
    """
    When called with the right finalizing phase event, will cause all identical Diff nodes to be merged with their size increasing.
    """
    from ...logger import logger
    from ..graph import Graph

    groups : dict[Diff, int] = {}

    def merge(u : Diff, v : Diff):
        """
        Merges vertex v into vertex u (v gets deleted at the end).
        """
        for e in v.edges:
            if e.source is v:
                e.source = u
            if e.destination is v:
                e.destination = u
        v.edges.clear()
        for g in Graph.active_graphs():
            g.remove(v)
        groups[u] += 1

    if e.major == "Finalizer" and e.minor == __N_diff_fusion_phase:
        logger.debug("Fusing {} Diff nodes.".format(len(Diff)))
        for u in Diff:
            merged = False
            for v in filter(lambda v : v in groups, Diff):
                if u.content == v.content:
                    merge(v, u)
                    merged = True
                    break
            if not merged:
                groups[u] = 1
        
        resize_work = [u for u, n in groups.items() if n > 1]
        if resize_work:
            logger.debug("Resizing {} fused Diff nodes. Actually lost {} Diff nodes.".format(len(resize_work), len(Diff) - len(groups)))
            for u in resize_work:
                u.size *= groups[u] ** 0.5

        
BuildingPhase.add_callback(finalize_diff_nodes)
BuildingPhase.add_callback(normalize_diff_nodes)
BuildingPhase.add_callback(fuse_diff_nodes)


del ABCMeta, Iterator, Optional, Color, Arrow, UniqueVertex, Vertex, chrono, logger, File, Connection, Handle, Socket, BuildingPhase, finalize_data_nodes, finalize_diff_nodes, normalize_diff_nodes, fuse_diff_nodes, CompilationParameters, Sequence, Hashable, TypeVar, S


"""
This module defines the Vertices and Edges that are related to execution.
"""

from pathlib import PurePath
from ...logger import logger
from typing import Any, Callable, Dict, Iterator, List, Optional, Set
from ..colors import Color
from Viper.meta.iterable import InstancePreservingClass
from ..graph import UniqueVertex, Arrow, Edge, Vertex
from ..utils import chrono
from ..build import BuildingPhase

__all__ = ["Process", "Thread", "Call", "CallHandler", "HasChildProcess", "UsesAsArgument", "HasThread", "HasFirstCall", "FollowedBy", "NextSignificantCall", "StartedProcess", "InjectedThread", "StartedThread"]





logger.debug("Loading execution library.")

CommandTree = list[tuple[str, "CommandTree"]]

class Process(UniqueVertex):

    """
    A process vertex. Holds information to identify a process.
    """

    __slots__ = {
        "PID" : "The PID of the process during its execution",
        "__command" : "The command that this process is running",
        "executable" : "The file system path that the process was started in",
        "start" : "The time at which the process was started",
        "stop" : "The time at chich the process was stopped",
        "__sub_commands" : "The tree of commands executed by all chile processes"
    }

    __pickle_slots__ = {
        "PID",
        "command",
        "executable",
        "start",
        "stop",
        "sub_commands"
    }

    def __init__(self, *, c: Color | None = None, parent: Optional[Vertex] = None) -> None:
        super().__init__(c=c, parent=parent)
        self.PID : int = 0
        self.__command : tuple[str] = tuple()
        self.executable : str = ""
        self.start : float = 0.0
        self.stop : float = 0.0
        self.__sub_commands : CommandTree | None = None

        from .filesystem import NewFile
        from argparse import ArgumentParser
        from pathlib import PurePath
        from ..utils import is_path, path_factory

        def to_absolute(p : PurePath, cwd : PurePath) -> PurePath:
            if p.is_absolute():
                return p
            return cwd / p
        
        p = ArgumentParser()
        
        def link(e : NewFile):
            from .filesystem import File
            f = e.file
            if not f.name:
                return
            # print("New file :", repr(f.path))
            # # print("Process :", repr(self.command))
            # print("Process #{} : '{}'".format(self.PID, self.executable))
            # print("Splitting command : {}".format(self.__command))
            for i, arg in enumerate(self.__command):
                if not i:
                    arg = self.executable
                while arg.endswith(" "):
                    arg = arg[:-1]
                while arg.startswith(" "):
                    arg = arg[1:]
                if arg.startswith("\"") and arg.endswith("\""):
                    arg = arg[1:-1]
                if arg.startswith("'") and arg.endswith("'"):
                    arg = arg[1:-1]
                # print(">>>", f.name.lower(), arg.lower())
                if (is_path(arg) and f.name.lower() == to_absolute(path_factory(arg), path_factory(self.executable)).name.lower()) or f.name.lower() in arg.lower():
                # if (f.name in arg and len(f.name) / len(arg) > 0.9) or (str(f.path) in arg and len(str(f.path)) / len(arg) > 0.9):        # You need to work with a process cwd
                    if i > 0:
                        UsesAsArgument(self, f)
                        break
                    else:
                        Runs(self, f)

        NewFile.add_callback(link)

    @property
    def label(self) -> str:
        """
        Returns a label for the Process node.
        """
        return "Process #" + str(self.PID)
        
    @property
    def threads(self) -> list["Thread"]:
        """
        List of all the threads that this process had.
        """
        return [e.destination for e in self.edges if isinstance(e, HasThread)] 
    
    @property
    def children_processes(self) -> list["Process"]:
        """
        List of all the children processes that this process created.
        """
        return [e.destination for e in self.edges if isinstance(e, HasChildProcess) and e.source is self]
    
    @property
    def parent_process(self) -> Optional["Process"]:
        """
        Returns the parent process node if one exists in the graph.
        """
        for e in self.edges:
            if isinstance(e, HasChildProcess) and e.destination is self:
                return e.source
            
    @property
    def command(self) -> str:
        """
        The command ran by this Process.
        """
        return " ".join(self.__command)
    
    @command.setter
    def command(self, cmd : str):
        from ..utils import parse_command_line
        if not isinstance(cmd, str):
            raise TypeError("Expected str, got " + repr(type(cmd).__name__))
        self.__command = tuple(parse_command_line(cmd))
        if not self.__command and self.executable:
            self.__command = (self.executable, )
        
    def parsed_command(self) -> list[str]:
        """
        Returns the argument vector used for starting this process.
        """
        return list(self.__command)

    @property
    def sub_commands(self) -> CommandTree:
        """
        Returns a dict structure that represents the commands executed by all child processes.
        """
        if self.__sub_commands != None:
            return self.__sub_commands
        sc = []
        for e in self.edges:
            if isinstance(e, HasChildProcess) and e.source is self:
                sc.append((e.destination.command, e.destination.sub_commands))
        self.__sub_commands = sc
        return sc
    
    @sub_commands.setter
    def sub_commands(self, value : CommandTree):
        self.__sub_commands = value





class Thread(UniqueVertex):

    """
    A thread vertex. Holds information to identify a thread.
    """

    __slots__ = {
        "TID" : "The TID of the thread during its execution",
        "Ncalls" : "The number of system calls that the thread made",
        "first" : "The first system call that this thread made",
        "last" : "the last system call that this thread made",
        "start" : "The time at which the thread was started",
        "stop" : "The time at chich the thread was stopped"
    }

    __pickle_slots__ = {
        "TID",
        "Ncalls",
        "first",
        "last",
        "start",
        "stop"
    }

    def __init__(self, *, c: Color | None = None, parent: Optional[Vertex] = None) -> None:
        super().__init__(c=c, parent=parent)
        self.TID : int = 0
        self.Ncalls : int = 0
        self.first : Optional["Call"] = None
        self.last : Optional["Call"] = None
        self.start : float = 0.0
        self.stop : float = 0.0

    @property
    def label(self) -> str:
        """
        Returns a label for the Thread node.
        """
        return "Thread #" + str(self.TID)
    
    @property
    def process(self) -> Process:
        """
        The Process Vertex that runs this Thread.
        """
        for e in self.edges:
            if isinstance(e, HasThread):
                return e.source
        raise RuntimeError("Got a Thread without a parent Process.")





class Call(UniqueVertex):

    """
    A system call vertex. Holds information about a specific system call.
    Don't forget to call c.integrate() after setup of Call c is finished to integrate the call in the graph.
    """

    __slots__ = {
        "name" : "The name of the system call",
        "category" : "The family of the system call",
        "stacktrace" : "The stacktrace of the call",
        "status" : "Indicates if the call ran succesfully",
        "arguments" : "The arguments that the call received",
        "return_value" : "The value that the call returned",
        "time" : "The timestamp at which the call was detected",
        "flags" : "The flags of the call",
        "__thread" : "A shortcut to the Thread vertex that made this call"
    }

    __pickle_slots__ = {
        "name",
        "category",
        "stacktrace",
        "status",
        "arguments",
        "return_value",
        "time",
        "flags",
        "thread"
    }

    def __init__(self, *, c: Color | None = None, parent: Optional[Vertex] = None) -> None:
        from ..record import record
        super().__init__(c=c, parent=parent)
        self.name : str = ""
        self.category : str = ""
        self.stacktrace : tuple = ()
        self.status : int = 0
        self.arguments : record = record()
        self.return_value : Any = None
        self.time : float = 0.0
        self.flags : record = record()
        self.__thread : Thread | None = None
    
    @property
    def thread(self) -> Thread:
        """
        The Thread that made this Call.
        """
        if self.__thread is None:
            raise RuntimeError("Got a Call without a calling Thread.")
        return self.__thread
    
    @thread.setter
    def thread(self, value : Thread):
        if not isinstance(value, Thread):
            raise TypeError("Expected Thread, got " + repr(type(value).__name__))
        self.__thread = value

    __seen : Set["CallHandler"] = set()             # The individual system call handlers already seen
    __cache : Dict[str, List["CallHandler"]] = {}   # Used to speed up access to the correct handlers for next calls to an already seen system call (with the same name)

    @chrono
    def integrate(self):
        """
        Integrates the call to the graph.
        Finds the correct handler(s) for this call and executes them.
        """
        if self.name not in Call.__cache:   # Add the system call name in cache if not present
            Call.__cache[self.name] = []
        if self.name in Call.__cache:       # Some handlers have already been matched
            for CH in Call.__cache[self.name]:
                if self in CH:
                    CH(self)
        if len(Call.__seen) < len(CallHandler):  # There might be unmatched handlers
            for CH in filter(lambda CH : CH not in Call.__seen, CallHandler):
                if self in CH:
                    Call.__cache[self.name].append(CH)
                    Call.__seen.add(CH)
                    CH(self)
    
    @property
    def label(self) -> str:
        """
        The label for this node.
        """
        return self.name





class CallHandler(metaclass = InstancePreservingClass):

    """
    The class for all handlers of system calls.
    Upon creation, a system call vertex will look into the instances of this class to find a valid handler and execute its integration function.
    """

    def __init__(self, matcher : Callable[[Call], bool], integrator : Callable[[Call], None]) -> None:
        self.__matcher = matcher
        self.__integrator = integrator
    

    def __repr__(self) -> str:
        """
        Implements repr(self).
        """
        return "CallHandler(" + repr(self.__matcher) + ", " + repr(self.__integrator) + ")"
    

    def match(self, c : Call) -> bool:
        """
        Returns True if given Call c should be handled by this CallHandler
        """
        if self.__matcher(c):
            return True
        return False
    

    def integrate(self, c : Call):
        """
        Integrates Call c into the graph
        """
        self.__integrator(c)
    

    def __contains__(self, c : Any) -> bool:
        """
        Implements c in self.
        Equivalent to self.match(c).
        """
        if not isinstance(c, Call):
            return False
        return self.match(c)
    
    
    def __call__(self, c : Call):
        """
        Implements self(c).
        Equivalent to self.integrate(c)
        """
        self.integrate(c)





class HasChildProcess(Arrow):

    """
    This kind of arrow indicates that a process created another one.
    """

    source : Process
    destination : Process





class UsesAsArgument(Edge):

    """
    This kind of edge indicates that a process used a file or directory in its command line.
    """

    from typing import TYPE_CHECKING
    if TYPE_CHECKING:
        from .filesystem import File, Directory
    del TYPE_CHECKING

    source : Process
    destination : "File | Directory"




class Runs(Edge):
    
    """
    This kind of edge indicates process ran a file as its program.
    """

    from typing import TYPE_CHECKING
    if TYPE_CHECKING:
        from .filesystem import File
    del TYPE_CHECKING

    source : Process
    destination : "File"





class HasThread(Edge):

    """
    This kind of edge indicates that a process hosts a thread.
    """

    label : str = ""

    source : Process
    destination : Thread





class HasFirstCall(Edge):

    """
    This kind of edge indicates what was the first system call of a thread.
    """

    label : str = ""

    source : Thread
    destination : Call





class FollowedBy(Arrow):

    """
    This kind of arrow indicates that a system call was followed by another one.
    """

    label : str = ""

    source : Call
    destination : Call





class NextSignificantCall(Arrow):
    
    """
    This kind of arrows links two Call nodes that happened one after the other in the same thread both nodes are linked to other types of nodes.
    """

    label : str = ""

    source : Call
    destination : Call





class StartedThread(Edge):

    """
    This kind of edge indicates that a system call started a new thread.
    """

    __slots__ = {
        "remote" : "Indicates if the started thread was started in another process (remote thread)"
    }

    __pickle_slots__ = {
        "remote"
    }

    label : str = ""

    source : Call
    destination : Thread

    def __init__(self, source: Vertex, destination: Vertex, *, c: Color | None = None, auto_write: bool = True) -> None:
        super().__init__(source, destination, c=c, auto_write=auto_write)
        self.remote : bool = False
    




class InjectedThread(Arrow):

    """
    This kind of arrow indicates a process created a remote thread.
    """

    source : Process
    destination : Thread





class StartedProcess(Edge):

    """
    This kind of edge indicates that a system call started a new process.
    """

    label : str = ""

    source : Call
    destination : Process





@chrono
def integrate_thread_creation(c : Call):
    """
    Links a Thread vertex to the Call vertex (from another thread) that created it.
    """
    from ...logger import logger
    from ..graph import find_or_create
    if c.arguments.thread_identifier != 0:      # Should indicate that the call failed
        logger.debug("New thread detected.")
        t : Thread
        t, ok = find_or_create(Thread, TID = c.arguments.thread_identifier)
        l = StartedThread(c, t)
        if c.name in {"CreateRemoteThread", "CreateRemoteThreadEx"}:
            logger.info("Detected thread injection.")
            l.remote = True
            p = c.thread.process
            InjectedThread(p, t)

@chrono
def integrate_process_creation(c : Call):
    """
    Links a Process vertex to the Call vertex (from another process) that created it.
    """
    from ...logger import logger
    from ..graph import find_or_create
    if c.arguments.process_identifier != 0:
        logger.debug("New process detected.")
        p : Process
        p, ok = find_or_create(Process, PID = c.arguments.process_identifier)
        l = StartedProcess(c, p)
        




# Thread creation handlers
CallHandler(lambda c : c.name == "CreateThread", integrate_thread_creation)          # Basic thread creation
CallHandler(lambda c : c.name in {"CreateRemoteThread", "CreateRemoteThreadEx"}, integrate_thread_creation)    # Remote thread creation (into another process)

# Process creation handlers
CallHandler(lambda c : "CreateProcess" in c.name, integrate_process_creation)


__N_phase = BuildingPhase.request_finalizing_phase()

@chrono
def remove_cuckoo_phantom_processes(ev : BuildingPhase):
    """
    When called with the right finalizing phase event, will cause Process nodes which represent Cuckoo process (inital processes with no info available) to be erased.
    """
    from ...logger import logger
    from ..graph import Graph
    from .network import SpawnedProcess, Host
    if ev.major == "Finalizer" and ev.minor == __N_phase:
        logger.debug("Erasing Cuckoo phantom processes.")
        n = 0
        h = Host.current
        for e in h.edges.copy():
            p = e.destination
            if isinstance(p, Process) and not p.parent_process and not p.executable:
                for pi in p.children_processes:
                    SpawnedProcess(h, pi)
                for ei in p.edges.copy():
                    if isinstance(e, SpawnedProcess):
                        h = e.source
                    ei.delete()
                    for g in Graph.active_graphs():
                        g.remove(ei)
                for g in Graph.active_graphs():
                    g.remove(p)
                n += 1
        if n:
            logger.debug("Actually removed {} Cuckoo process{}.".format(n, "es" if n > 1 else ""))



BuildingPhase.add_callback(remove_cuckoo_phantom_processes)

del logger, Any, Callable, Dict, Iterator, List, Optional, Set, Color, InstancePreservingClass, UniqueVertex, Arrow, Vertex, PurePath, chrono, integrate_process_creation, integrate_thread_creation, remove_cuckoo_phantom_processes